@@ -1,4 +1,5 @@
 #include "filezilla.h"
+#include "activity_logger_layer.h"
 #include "controlsocket.h"
 #include "directorycache.h"
 #include "engineprivate.h"
@@ -145,10 +146,11 @@ int CControlSocket::ResetOperation(int nErrorCode)
 			nErrorCode == FZ_REPLY_ERROR_NOTFOUND)
 		{
 			if (!oldOperation->topLevelOperation_) {
-				return ParseSubcommandResult(nErrorCode, *oldOperation);
+				return ParseSubcommandResult(nErrorCode, std::move(oldOperation));
 			}
 		}
 		else {
+			oldOperation.reset();
 			return ResetOperation(nErrorCode);
 		}
 	}
@@ -213,6 +215,8 @@ int CControlSocket::ResetOperation(int nErrorCode)
 			}
 			break;
 		}
+
+		oldOperation.reset();
 	}
 
 	engine_.transfer_status_.Reset();
@@ -249,16 +253,21 @@ int CControlSocket::DoClose(int nErrorCode)
 std::wstring CControlSocket::ConvertDomainName(std::wstring const& domain)
 {
 #ifdef FZ_WINDOWS
-	int len = IdnToAscii(IDN_ALLOW_UNASSIGNED, domain.c_str(), domain.size() + 1, nullptr, 0);
+	if (domain.size() > 1204 * 1024) {
+		return domain;
+	}
+
+	int const inlen = static_cast<int>(domain.size() + 1);
+	int len = IdnToAscii(IDN_ALLOW_UNASSIGNED, domain.c_str(), inlen, nullptr, 0);
 	if (!len) {
 		log(logmsg::debug_warning, L"Could not convert domain name");
 		return domain;
 	}
 
 	wchar_t* output = new wchar_t[len];
-	int res = IdnToAscii(IDN_ALLOW_UNASSIGNED, domain.c_str(), domain.size() + 1, output, len);
+	int res = IdnToAscii(IDN_ALLOW_UNASSIGNED, domain.c_str(), inlen, output, len);
 	if (!res) {
-		delete [] output;
+		delete[] output;
 		log(logmsg::debug_warning, L"Could not convert domain name");
 		return domain;
 	}
@@ -361,13 +370,15 @@ int CControlSocket::CheckOverwriteFile()
 	}
 
 	auto & data = static_cast<CFileTransferOpData &>(*operations_.back());
+	data.localFileSize_ = data.download() ? data.writer_factory_.size() : data.reader_factory_.size();
+	data.localFileTime_ = data.download() ? data.writer_factory_.mtime() : data.reader_factory_.mtime();
 
 	if (data.download()) {
-		if (fz::local_filesys::get_file_type(fz::to_native(data.localFile_), true) != fz::local_filesys::file) {
+		if (data.localFileSize_ == aio_base::nosize && data.localFileTime_.empty()) {
 			return FZ_REPLY_OK;
 		}
 	}
-
+	
 	CDirentry entry;
 	bool dirDidExist;
 	bool matchedCase;
@@ -381,24 +392,34 @@ int CControlSocket::CheckOverwriteFile()
 	bool found = engine_.GetDirectoryCache().LookupFile(entry, currentServer_, remotePath, data.remoteFile_, dirDidExist, matchedCase);
 
 	// Ignore entries with wrong case
-	if (found && !matchedCase)
+	if (found && !matchedCase) {
 		found = false;
+	}
+
+	if (found) {
+		if (data.remoteFileTime_.empty() && entry.has_date()) {
+			data.remoteFileTime_ = entry.time;
+		}
+	}
 
 	if (!data.download()) {
-		if (!found && data.remoteFileSize_ < 0 && data.fileTime_.empty()) {
+		if (!found && data.remoteFileSize_ < 0 && data.remoteFileTime_.empty()) {
 			return FZ_REPLY_OK;
 		}
 	}
 
+	
+
 	auto notification = std::make_unique<CFileExistsNotification>();
 
 	notification->download = data.download();
-	notification->localFile = data.localFile_;
+	notification->localFile = data.localName_;
 	notification->remoteFile = data.remoteFile_;
 	notification->remotePath = data.remotePath_;
-	notification->localSize = data.localFileSize_;
+	notification->localSize = static_cast<int64_t>(data.localFileSize_);
+	notification->localTime = data.localFileTime_;
 	notification->remoteSize = data.remoteFileSize_;
-	notification->remoteTime = data.fileTime_;
+	notification->remoteTime = data.remoteFileTime_;
 
 	if (currentServer_.HasFeature(ProtocolFeature::DataTypeConcept)) {
 		notification->ascii = data.flags_ & ftp_transfer_flags::ascii;
@@ -414,15 +435,6 @@ int CControlSocket::CheckOverwriteFile()
 		notification->canResume = false;
 	}
 
-	notification->localTime = fz::local_filesys::get_modification_time(fz::to_native(data.localFile_));
-
-	if (found) {
-		if (notification->remoteTime.empty() && entry.has_date()) {
-			notification->remoteTime = entry.time;
-			data.fileTime_ = entry.time;
-		}
-	}
-
 	SendAsyncRequest(std::move(notification));
 
 	return FZ_REPLY_WOULDBLOCK;
@@ -442,11 +454,13 @@ void SleepOpData::operator()(fz::event_base const&)
 	controlSocket_.ResetOperation(FZ_REPLY_OK);
 }
 
-CFileTransferOpData::CFileTransferOpData(wchar_t const* name, std::wstring const& local_file, std::wstring const& remote_file, CServerPath const& remote_path, transfer_flags const& flags)
+CFileTransferOpData::CFileTransferOpData(wchar_t const* name, CFileTransferCommand const& cmd)
 	: COpData(Command::transfer, name)
-	, flags_(flags)
-	, localFile_(local_file), remoteFile_(remote_file), remotePath_(remote_path)
+	, flags_(cmd.GetFlags())
+	, reader_factory_(cmd.GetReader()), writer_factory_(cmd.GetWriter()), localName_(reader_factory_ ? reader_factory_.name() : writer_factory_.name()), remoteFile_(cmd.GetRemoteFile()), remotePath_(cmd.GetRemotePath())
 {
+	localFileSize_ = download() ? writer_factory_.size() : reader_factory_.size();
+	localFileTime_ = download() ? writer_factory_.mtime() : reader_factory_.mtime();
 }
 
 std::wstring CControlSocket::ConvToLocal(char const* buffer, size_t len)
@@ -607,17 +621,21 @@ int CControlSocket::SendNextCommand()
 	return FZ_REPLY_OK;
 }
 
-int CControlSocket::ParseSubcommandResult(int prevResult, COpData const& opData)
+int CControlSocket::ParseSubcommandResult(int prevResult, std::unique_ptr<COpData> && previousOperation)
 {
 	if (operations_.empty()) {
 		log(logmsg::debug_warning, L"CControlSocket::ParseSubcommandResult(%d) called without active operation", prevResult);
+		previousOperation.reset();
 		ResetOperation(FZ_REPLY_ERROR);
 		return FZ_REPLY_ERROR;
 	}
 
 	auto & data = *operations_.back();
 	log(logmsg::debug_verbose, L"%s::SubcommandResult(%d) in state %d", data.name_, prevResult, data.opState);
-	int res = data.SubcommandResult(prevResult, opData);
+	int res = data.SubcommandResult(prevResult, *previousOperation);
+
+	previousOperation.reset();
+
 	if (res == FZ_REPLY_WOULDBLOCK) {
 		return FZ_REPLY_WOULDBLOCK;
 	}
@@ -696,11 +714,6 @@ CRealControlSocket::~CRealControlSocket()
 	ResetSocket();
 }
 
-bool CRealControlSocket::Connected() const
-{
-	return socket_ ? socket_->is_connected() : false;
-}
-
 int CRealControlSocket::Send(unsigned char const* buffer, unsigned int len)
 {
 	if (!active_layer_) {
@@ -725,7 +738,7 @@ int CRealControlSocket::Send(unsigned char const* buffer, unsigned int len)
 		}
 
 		if (written) {
-			SetActive(CFileZillaEngine::send);
+			RecordActivity(activity_logger::send, written);
 		}
 
 		if (static_cast<unsigned int>(written) < len) {
@@ -826,7 +839,7 @@ int CRealControlSocket::OnSend()
 		}
 
 		if (written) {
-			SetActive(CFileZillaEngine::send);
+			RecordActivity(activity_logger::send, written);
 
 			send_buffer_.consume(static_cast<size_t>(written));
 		}
@@ -857,7 +870,8 @@ int CRealControlSocket::DoConnect(std::wstring const& host, unsigned int port)
 
 	ResetSocket();
 	socket_ = std::make_unique<fz::socket>(engine_.GetThreadPool(), nullptr);
-	ratelimit_layer_ = std::make_unique<fz::rate_limited_layer>(this, *socket_, &engine_.GetRateLimiter());
+	activity_logger_layer_ = std::make_unique<activity_logger_layer>(nullptr, *socket_, engine_.activity_logger_);
+	ratelimit_layer_ = std::make_unique<fz::rate_limited_layer>(this, *activity_logger_layer_, &engine_.GetRateLimiter());
 	active_layer_ = ratelimit_layer_.get();
 
 	const int proxy_type = engine_.GetOptions().get_int(OPTION_PROXY_TYPE);
@@ -882,6 +896,8 @@ int CRealControlSocket::DoConnect(std::wstring const& host, unsigned int port)
 		}
 	}
 
+	SetSocketBufferSizes();
+
 	int res = active_layer_->connect(fz::to_native(ConvertDomainName(host)), port);
 
 	if (res) {
@@ -907,6 +923,7 @@ void CRealControlSocket::ResetSocket()
 	// Destroy in reverse order
 	proxy_layer_.reset();
 	ratelimit_layer_.reset();
+	activity_logger_layer_.reset();
 	socket_.reset();
 
 	send_buffer_.clear();
@@ -946,7 +963,7 @@ bool CControlSocket::SetFileExistsAction(CFileExistsNotification *pFileExistsNot
 				log(logmsg::status, _("Skipping download of %s"), filename);
 			}
 			else {
-				log(logmsg::status, _("Skipping upload of %s"), data.localFile_);
+				log(logmsg::status, _("Skipping upload of %s"), data.localName_);
 			}
 			ResetOperation(FZ_REPLY_OK);
 		}
@@ -963,7 +980,7 @@ bool CControlSocket::SetFileExistsAction(CFileExistsNotification *pFileExistsNot
 				log(logmsg::status, _("Skipping download of %s"), filename);
 			}
 			else {
-				log(logmsg::status, _("Skipping upload of %s"), data.localFile_);
+				log(logmsg::status, _("Skipping upload of %s"), data.localName_);
 			}
 			ResetOperation(FZ_REPLY_OK);
 		}
@@ -989,13 +1006,13 @@ bool CControlSocket::SetFileExistsAction(CFileExistsNotification *pFileExistsNot
 				log(logmsg::status, _("Skipping download of %s"), filename);
 			}
 			else {
-				log(logmsg::status, _("Skipping upload of %s"), data.localFile_);
+				log(logmsg::status, _("Skipping upload of %s"), data.localName_);
 			}
 			ResetOperation(FZ_REPLY_OK);
 		}
 		break;
 	case CFileExistsNotification::resume:
-		if (data.download() && data.localFileSize_ >= 0) {
+		if (data.download() && data.localFileSize_ != aio_base::nosize) {
 			data.resume_ = true;
 		}
 		else if (!data.download() && data.remoteFileSize_ >= 0) {
@@ -1006,32 +1023,17 @@ bool CControlSocket::SetFileExistsAction(CFileExistsNotification *pFileExistsNot
 	case CFileExistsNotification::rename:
 		if (data.download()) {
 			{
-				std::wstring tmp;
-				CLocalPath l(data.localFile_, &tmp);
-				if (l.empty() || tmp.empty()) {
-					ResetOperation(FZ_REPLY_INTERNALERROR);
-					return false;
-				}
-				if (!l.ChangePath(pFileExistsNotification->newName)) {
-					ResetOperation(FZ_REPLY_INTERNALERROR);
-					return false;
-				}
-				if (!l.HasParent() || !l.MakeParent(&tmp)) {
+				if (!pFileExistsNotification->new_writer_factory_) {
 					ResetOperation(FZ_REPLY_INTERNALERROR);
 					return false;
 				}
 
-				data.localFile_ = l.GetPath() + tmp;
+				data.writer_factory_ = std::move(pFileExistsNotification->new_writer_factory_);
+				data.localName_ = data.writer_factory_.name();
 			}
 
-			int64_t size;
-			bool isLink;
-			if (fz::local_filesys::get_file_info(fz::to_native(data.localFile_), isLink, &size, nullptr, nullptr) == fz::local_filesys::file) {
-				data.localFileSize_ = size;
-			}
-			else {
-				data.localFileSize_ = -1;
-			}
+			data.localFileSize_ = data.writer_factory_.size();
+			data.localFileTime_ = data.writer_factory_.mtime();
 
 			if (CheckOverwriteFile() == FZ_REPLY_OK) {
 				SendNextCommand();
@@ -1039,8 +1041,8 @@ bool CControlSocket::SetFileExistsAction(CFileExistsNotification *pFileExistsNot
 		}
 		else {
 			data.remoteFile_ = pFileExistsNotification->newName;
-			data.fileTime_ = fz::datetime();
 			data.remoteFileSize_ = -1;
+			data.remoteFileTime_ = fz::datetime();
 
 			CDirentry entry;
 			bool dir_did_exist;
@@ -1050,7 +1052,7 @@ bool CControlSocket::SetFileExistsAction(CFileExistsNotification *pFileExistsNot
 			{
 				data.remoteFileSize_ = entry.size;
 				if (entry.has_date()) {
-					data.fileTime_ = entry.time;
+					data.remoteFileTime_ = entry.time;
 				}
 
 				if (CheckOverwriteFile() != FZ_REPLY_OK) {
@@ -1067,7 +1069,7 @@ bool CControlSocket::SetFileExistsAction(CFileExistsNotification *pFileExistsNot
 			log(logmsg::status, _("Skipping download of %s"), filename);
 		}
 		else {
-			log(logmsg::status, _("Skipping upload of %s"), data.localFile_);
+			log(logmsg::status, _("Skipping upload of %s"), data.localName_);
 		}
 		ResetOperation(FZ_REPLY_OK);
 		break;
@@ -1080,26 +1082,6 @@ bool CControlSocket::SetFileExistsAction(CFileExistsNotification *pFileExistsNot
 	return true;
 }
 
-void CControlSocket::CreateLocalDir(std::wstring const& local_file)
-{
-	std::wstring file;
-	CLocalPath local_path(local_file, &file);
-	if (local_path.empty() || !local_path.HasParent()) {
-		return;
-	}
-
-	fz::native_string last_created;
-	fz::mkdir(fz::to_native(local_path.GetPath()), true, false, &last_created);
-
-	if (!last_created.empty()) {
-		// Send out notification
-		auto n = std::make_unique<CLocalDirCreatedNotification>();
-		if (n->dir.SetPath(fz::to_wstring(last_created))) {
-			engine_.AddNotification(std::move(n));
-		}
-	}
-}
-
 void CControlSocket::List(CServerPath const&, std::wstring const&, int)
 {
 	Push(std::make_unique<CNotSupportedOpData>());
@@ -1152,10 +1134,10 @@ void CControlSocket::operator()(fz::event_base const& ev)
 		&CControlSocket::OnObtainLock);
 }
 
-void CControlSocket::SetActive(CFileZillaEngine::_direction direction)
+void CControlSocket::RecordActivity(activity_logger::_direction direction, uint64_t amount)
 {
 	SetAlive();
-	engine_.SetActive(direction);
+	engine_.activity_logger_.record(direction, amount);
 }
 
 void CControlSocket::SendDirectoryListingNotification(CServerPath const& path, bool failed)
