@@ -17,10 +17,10 @@
 #include "../directorycache.h"
 #include "../directorylistingparser.h"
 #include "../engineprivate.h"
-#include "../iothread.h"
 #include "../pathcache.h"
 #include "../proxy.h"
 #include "../servercapabilities.h"
+#include "../tls.h"
 
 #include "../../include/externalipresolver.h"
 #include "../../include/engine_options.h"
@@ -28,7 +28,6 @@
 #include <libfilezilla/file.hpp>
 #include <libfilezilla/iputils.hpp>
 #include <libfilezilla/local_filesys.hpp>
-#include <libfilezilla/tls_layer.hpp>
 #include <libfilezilla/util.hpp>
 
 #include <algorithm>
@@ -57,7 +56,7 @@ void CFtpControlSocket::OnReceive()
 		int error;
 
 		size_t const toRead = max - receiveBuffer_.size();
-		int read = active_layer_->read(receiveBuffer_.get(toRead), toRead, error);
+		int read = active_layer_->read(receiveBuffer_.get(toRead), static_cast<unsigned int>(toRead), error);
 		if (read < 0) {
 			if (error != EAGAIN) {
 				log(logmsg::error, _("Could not read from socket: %s"), fz::socket_error_description(error));
@@ -79,7 +78,7 @@ void CFtpControlSocket::OnReceive()
 		size_t i = receiveBuffer_.size();
 		receiveBuffer_.add(read);
 
-		SetActive(CFileZillaEngine::recv);
+		SetAlive();
 
 		while (i < receiveBuffer_.size()) {
 			auto const& p = receiveBuffer_[i];
@@ -123,7 +122,7 @@ void CFtpControlSocket::ParseLine(std::wstring line)
 
 	if (!operations_.empty() && operations_.back()->opId == Command::connect) {
 		auto & data = static_cast<CFtpLogonOpData &>(*operations_.back());
-		if (data.waitChallenge) {
+		if (data.challengeMode_) {
 			std::wstring& challenge = data.challenge;
 			if (!challenge.empty())
 #ifdef FZ_WINDOWS
@@ -180,7 +179,6 @@ void CFtpControlSocket::OnConnect()
 {
 	m_lastTypeBinary = -1;
 	m_sentRestartOffset = false;
-	m_protectDataChannel = false;
 
 	SetAlive();
 
@@ -191,6 +189,8 @@ void CFtpControlSocket::OnConnect()
 			tls_layer_ = std::make_unique<fz::tls_layer>(event_loop_, this, *active_layer_, &engine_.GetContext().GetTlsSystemTrustStore(), logger_);
 			active_layer_ = tls_layer_.get();
 
+			tls_layer_->set_alpn("ftp");
+			tls_layer_->set_min_tls_ver(get_min_tls_ver(engine_.GetOptions()));
 			if (!tls_layer_->client_handshake(this)) {
 				DoClose();
 			}
@@ -346,18 +346,6 @@ int CFtpControlSocket::ResetOperation(int nErrorCode)
 				}
 			}
 		}
-		if (nErrorCode != FZ_REPLY_OK && data.download() && !data.fileDidExist_) {
-			data.ioThread_.reset();
-			int64_t size;
-			bool isLink;
-			if (fz::local_filesys::get_file_info(fz::to_native(data.localFile_), isLink, &size, nullptr, nullptr) == fz::local_filesys::file && size == 0) {
-				// Download failed and a new local file was created before, but
-				// nothing has been written to it. Remove it again, so we don't
-				// leave a bunch of empty files all over the place.
-				log(logmsg::debug_verbose, L"Deleting empty file");
-				fz::remove_file(fz::to_native(data.localFile_));
-			}
-		}
 	}
 
 	if (!operations_.empty() && operations_.back()->opId == PrivCommand::rawtransfer &&
@@ -416,12 +404,11 @@ void CFtpControlSocket::ChangeDir(CServerPath const& path, std::wstring const& s
 	Push(std::move(pData));
 }
 
-void CFtpControlSocket::FileTransfer(std::wstring const& localFile, CServerPath const& remotePath,
-									std::wstring const& remoteFile, transfer_flags const& flags)
+void CFtpControlSocket::FileTransfer(CFileTransferCommand const& cmd)
 {
 	log(logmsg::debug_verbose, L"CFtpControlSocket::FileTransfer()");
 
-	auto pData = std::make_unique<CFtpFileTransferOpData>(*this, localFile, remoteFile, remotePath, flags);
+	auto pData = std::make_unique<CFtpFileTransferOpData>(*this, cmd);
 	Push(std::move(pData));
 }
 
@@ -453,6 +440,12 @@ void CFtpControlSocket::TransferEnd()
 		data.pOldData->transferEndReason = reason;
 	}
 
+	if (reason == TransferEndReason::failed_tls_resumption) {
+		log(logmsg::error, _("TLS session resumption on data connection failed. Closing control connection to start over."));
+		DoClose(FZ_REPLY_ERROR | FZ_REPLY_DISCONNECTED);
+		return;
+	}
+
 	switch (data.opState)
 	{
 	case rawtransfer_transfer:
@@ -501,6 +494,8 @@ bool CFtpControlSocket::SetAsyncRequestReply(CAsyncRequestNotification *pNotific
 				return false;
 			}
 			credentials_.SetPass(pInteractiveLoginNotification->credentials.GetPass());
+			credentials_.SetExtraParameters(currentServer_.GetProtocol(), pInteractiveLoginNotification->credentials.GetExtraParameters());
+
 			SendNextCommand();
 		}
 		break;
@@ -519,10 +514,8 @@ bool CFtpControlSocket::SetAsyncRequestReply(CAsyncRequestNotification *pNotific
 				return false;
 			}
 
-			if (!operations_.empty() && operations_.back()->opId == Command::connect &&
-				operations_.back()->opState == LOGON_AUTH_WAIT)
-			{
-				operations_.back()->opState = LOGON_LOGON;
+			if (!operations_.empty() && operations_.back()->opId == Command::connect) {
+				static_cast<CFtpLogonOpData&>(*operations_.back()).tls_handshake_finished();
 			}
 		}
 		break;
@@ -538,6 +531,21 @@ bool CFtpControlSocket::SetAsyncRequestReply(CAsyncRequestNotification *pNotific
 				return true;
 			}
 		}
+	case reqId_tls_no_resumption:
+		{
+			auto & notification = static_cast<FtpTlsNoResumptionNotification&>(*pNotification);
+			if (!notification.allow_) {
+				ResetOperation(FZ_REPLY_CANCELED);
+				return false;
+			}
+			else {
+				CServerCapabilities::SetCapability(currentServer_, tls_resumption, no);
+				if (!operations_.empty() && operations_.back()->opId == PrivCommand::rawtransfer && m_pTransferSocket) {
+					m_pTransferSocket->ContinueWithoutSesssionResumption();
+				}
+				return true;
+			}
+		}
 	default:
 		log(logmsg::debug_warning, L"Unknown request %d", pNotification->GetRequestID());
 		ResetOperation(FZ_REPLY_INTERNALERROR);
@@ -798,6 +806,7 @@ void CFtpControlSocket::ResetSocket()
 	m_Response.clear();
 	m_MultilineResponseCode.clear();;
 	m_MultilineResponseLines.clear();
+	m_protectDataChannel = false;
 
 	CRealControlSocket::ResetSocket();
 }
