@@ -7,7 +7,7 @@
 #include "event.h"
 #include "filetransfer.h"
 #include "list.h"
-#include "input_thread.h"
+#include "input_parser.h"
 #include "mkd.h"
 #include "rename.h"
 #include "rmd.h"
@@ -29,6 +29,10 @@
 
 #include <assert.h>
 
+#ifndef FZ_WINDOWS
+#include <unistd.h>
+#endif
+
 struct SftpRateAvailableEventType;
 typedef fz::simple_event<SftpRateAvailableEventType, fz::direction::type> SftpRateAvailableEvent;
 
@@ -64,7 +68,7 @@ void CSftpControlSocket::OnSftpEvent(sftp_message const& message)
 		return;
 	}
 
-	if (!input_thread_) {
+	if (!input_parser_) {
 		return;
 	}
 
@@ -102,10 +106,10 @@ void CSftpControlSocket::OnSftpEvent(sftp_message const& message)
 		log_raw(logmsg::status, message.text[0]);
 		break;
 	case sftpEvent::Recv:
-		SetActive(CFileZillaEngine::recv);
+		RecordActivity(activity_logger::recv, fz::to_integral<uint64_t>(message.text[0]));
 		break;
 	case sftpEvent::Send:
-		SetActive(CFileZillaEngine::send);
+		RecordActivity(activity_logger::send, fz::to_integral<uint64_t>(message.text[0]));
 		break;
 	case sftpEvent::Transfer:
 		{
@@ -113,23 +117,26 @@ void CSftpControlSocket::OnSftpEvent(sftp_message const& message)
 
 			bool tmp;
 			CTransferStatus status = engine_.transfer_status_.Get(tmp);
-			if (!status.empty() && !status.madeProgress) {
+			if (!status.empty()) {
 				if (!operations_.empty() && operations_.back()->opId == Command::transfer) {
 					auto & data = static_cast<CSftpFileTransferOpData &>(*operations_.back());
-					if (data.download()) {
-						if (value > 0) {
-							engine_.transfer_status_.SetMadeProgress();
-						}
+					if (!data.download()) {
+						engine_.transfer_status_.Update(value);	
 					}
-					else {
-						if (status.currentOffset > status.startOffset + 65565) {
-							engine_.transfer_status_.SetMadeProgress();
+					if (!status.madeProgress) {
+						if (data.download()) {
+							if (value > 0) {
+								engine_.transfer_status_.SetMadeProgress();
+							}
+						}
+						else {
+							if (status.currentOffset > status.startOffset + 65565) {
+								engine_.transfer_status_.SetMadeProgress();
+							}
 						}
 					}
 				}
 			}
-
-			engine_.transfer_status_.Update(value);
 		}
 		break;
 	case sftpEvent::AskHostkey:
@@ -239,11 +246,7 @@ void CSftpControlSocket::OnSftpEvent(sftp_message const& message)
 		{
 			auto tokens = fz::strtok_view(message.text[0], ' ');
 			if (!tokens.empty()) {
-				m_sftpEncryptionDetails.hostKeyFingerprintSHA256 = tokens.back();
-				tokens.pop_back();
-			}
-			if (!tokens.empty()) {
-				m_sftpEncryptionDetails.hostKeyFingerprintMD5 = tokens.back();
+				m_sftpEncryptionDetails.hostKeyFingerprint = tokens.back();
 				tokens.pop_back();
 			}
 			for (auto const& token : tokens) {
@@ -254,6 +257,30 @@ void CSftpControlSocket::OnSftpEvent(sftp_message const& message)
 			}
 		}
 		break;
+	case sftpEvent::io_nextbuf:
+		if (!operations_.empty() && operations_.back()->opId == Command::transfer) {
+			auto & data = static_cast<CSftpFileTransferOpData&>(*operations_.back());
+			data.OnNextBufferRequested(fz::to_integral<uint64_t>(message.text[0]));
+		}
+		break;
+	case sftpEvent::io_open:
+		if (!operations_.empty() && operations_.back()->opId == Command::transfer) {
+			auto & data = static_cast<CSftpFileTransferOpData&>(*operations_.back());
+			data.OnOpenRequested(fz::to_integral<uint64_t>(message.text[0]));
+		}
+		break;
+	case sftpEvent::io_size:
+		if (!operations_.empty() && operations_.back()->opId == Command::transfer) {
+			auto & data = static_cast<CSftpFileTransferOpData&>(*operations_.back());
+			data.OnSizeRequested();
+		}
+		break;
+	case sftpEvent::io_finalize:
+		if (!operations_.empty() && operations_.back()->opId == Command::transfer) {
+			auto & data = static_cast<CSftpFileTransferOpData&>(*operations_.back());
+			data.OnFinalizeRequested(fz::to_integral<uint64_t>(message.text[0]));
+		}
+		break;
 	default:
 		log(logmsg::debug_warning, L"Message type %d not handled", message.type);
 		break;
@@ -266,7 +293,7 @@ void CSftpControlSocket::OnSftpListEvent(sftp_list_message const& message)
 		return;
 	}
 
-	if (!input_thread_) {
+	if (!input_parser_) {
 		return;
 	}
 
@@ -282,19 +309,6 @@ void CSftpControlSocket::OnSftpListEvent(sftp_list_message const& message)
 	}
 }
 
-void CSftpControlSocket::OnTerminate(std::wstring const& error)
-{
-	if (!error.empty()) {
-		log_raw(logmsg::error, error);
-	}
-	else {
-		log_raw(logmsg::debug_info, L"CSftpControlSocket::OnTerminate without error");
-	}
-	if (process_) {
-		DoClose();
-	}
-}
-
 int CSftpControlSocket::SendCommand(std::wstring const& cmd, std::wstring const& show)
 {
 	SetWait(true);
@@ -310,10 +324,10 @@ int CSftpControlSocket::SendCommand(std::wstring const& cmd, std::wstring const&
 		return FZ_REPLY_INTERNALERROR;
 	}
 
-	return AddToStream(cmd + L"\n");
+	return AddToSendBuffer(cmd + L"\n");
 }
 
-int CSftpControlSocket::AddToStream(std::wstring const& cmd)
+int CSftpControlSocket::AddToSendBuffer(std::wstring const& cmd)
 {
 	std::string const str = ConvToServer(cmd);
 	if (str.empty()) {
@@ -321,17 +335,25 @@ int CSftpControlSocket::AddToStream(std::wstring const& cmd)
 		return FZ_REPLY_ERROR;
 	}
 
-	return AddToStream(str);
+	return AddToSendBuffer(str);
 }
 
-int CSftpControlSocket::AddToStream(std::string const& cmd)
+int CSftpControlSocket::AddToSendBuffer(std::string const& cmd)
 {
 	if (!process_) {
 		return FZ_REPLY_INTERNALERROR;
 	}
 
-	if (!process_->write(cmd)) {
-		return FZ_REPLY_ERROR | FZ_REPLY_DISCONNECTED;
+
+	bool const can_send = send_buffer_.empty();
+
+	send_buffer_.append(cmd);
+
+	if (can_send) {
+		return SendToProcess();
+	}
+	else {
+		return FZ_REPLY_WOULDBLOCK;
 	}
 
 	return FZ_REPLY_WOULDBLOCK;
@@ -474,29 +496,26 @@ void CSftpControlSocket::ProcessReply(int result, std::wstring const& reply)
 		}
 	}
 }
-void CSftpControlSocket::FileTransfer(std::wstring const& localFile, CServerPath const& remotePath,
-									std::wstring const& remoteFile, transfer_flags const& flags)
+void CSftpControlSocket::FileTransfer(CFileTransferCommand const& cmd)
 {
-	auto pData = std::make_unique<CSftpFileTransferOpData>(*this, localFile, remoteFile, remotePath, flags);
-	Push(std::move(pData));
+	Push(std::make_unique<CSftpFileTransferOpData>(*this, cmd));
 }
 
 int CSftpControlSocket::DoClose(int nErrorCode)
 {
 	remove_bucket();
-
 	if (process_) {
 		process_->kill();
 	}
 
-	if (input_thread_) {
-		input_thread_.reset();
+	if (input_parser_) {
+		input_parser_.reset();
 
 		auto threadEventsFilter = [&](fz::event_loop::Events::value_type const& ev) -> bool {
 			if (ev.first != this) {
 				return false;
 			}
-			else if (ev.second->derived_type() == CSftpEvent::type() || ev.second->derived_type() == CTerminateEvent::type()) {
+			else if (ev.second->derived_type() == CSftpEvent::type() || ev.second->derived_type() == CSftpListEvent::type()) {
 				return true;
 			}
 			return false;
@@ -506,6 +525,13 @@ int CSftpControlSocket::DoClose(int nErrorCode)
 	}
 	process_.reset();
 
+#ifndef FZ_WINDOWS
+	if (shm_fd_ != -1) {
+		close(shm_fd_);
+		shm_fd_ = -1;
+	}
+#endif
+
 	m_sftpEncryptionDetails = CSftpEncryptionNotification();
 
 	return CControlSocket::DoClose(nErrorCode);
@@ -576,7 +602,7 @@ void CSftpControlSocket::OnQuotaRequest(fz::direction::type const d)
 
 	fz::rate::type bytes = available(d);
 	if (bytes == fz::rate::unlimited) {
-		AddToStream(fz::sprintf("-%d-\n", d));
+		AddToSendBuffer(fz::sprintf("-%d-\n", d));
 	}
 	else if (bytes > 0) {
 		int b;
@@ -586,17 +612,17 @@ void CSftpControlSocket::OnQuotaRequest(fz::direction::type const d)
 		else {
 			b = static_cast<int>(bytes);
 		}
-		AddToStream(fz::sprintf("-%d%d,%d\n", d, b, engine_.GetOptions().get_int(d ? OPTION_SPEEDLIMIT_OUTBOUND : OPTION_SPEEDLIMIT_INBOUND)));
+		AddToSendBuffer(fz::sprintf("-%d%d,%d\n", d, b, engine_.GetOptions().get_int(d ? OPTION_SPEEDLIMIT_OUTBOUND : OPTION_SPEEDLIMIT_INBOUND)));
 		consume(d, static_cast<fz::rate::type>(b));
 	}
 }
 
 void CSftpControlSocket::operator()(fz::event_base const& ev)
 {
-	if (fz::dispatch<CSftpEvent, CSftpListEvent, CTerminateEvent, SftpRateAvailableEvent>(ev, this,
+	if (fz::dispatch<fz::process_event, CSftpEvent, CSftpListEvent, SftpRateAvailableEvent>(ev, this,
+		&CSftpControlSocket::OnProcessEvent,
 		&CSftpControlSocket::OnSftpEvent,
 		&CSftpControlSocket::OnSftpListEvent,
-		&CSftpControlSocket::OnTerminate,
 		&CSftpControlSocket::OnQuotaRequest)) {
 		return;
 	}
@@ -615,3 +641,40 @@ void CSftpControlSocket::Push(std::unique_ptr<COpData> && pNewOpData)
 		}
 	}
 }
+
+int CSftpControlSocket::SendToProcess()
+{
+	if (!process_) {
+		return FZ_REPLY_INTERNALERROR;
+	}
+	while (!send_buffer_.empty()) {
+		fz::rwresult res = process_->write(send_buffer_.get(), send_buffer_.size());
+		if (res) {
+			send_buffer_.consume(res.value_);
+		}
+		else if (res.error_ == fz::rwresult::wouldblock) {
+			break;
+		}
+		else {
+			log(logmsg::error, _("Could not send command to fzsftp executable"));
+			return FZ_REPLY_ERROR | FZ_REPLY_DISCONNECTED;
+		}
+	}
+	return FZ_REPLY_WOULDBLOCK;
+}
+
+void CSftpControlSocket::OnProcessEvent(fz::process*, fz::process_event_flag const& f)
+{
+	if (f == fz::process_event_flag::write) {
+		int res = SendToProcess();
+		if (res != FZ_REPLY_WOULDBLOCK) {
+			DoClose(res);
+		}
+	}
+	else {
+		int res = input_parser_->OnData();
+		if (res != FZ_REPLY_WOULDBLOCK) {
+			DoClose(res);
+		}
+	}
+}
