@@ -4,12 +4,9 @@
 #include "../controlsocket.h"
 
 #include <libfilezilla/rate_limiter.hpp>
+#include <libfilezilla/process.hpp>
 
-namespace fz {
-class process;
-}
-
-class CSftpInputThread;
+class SftpInputParser;
 struct sftp_message;
 struct sftp_list_message;
 
@@ -22,8 +19,7 @@ public:
 	virtual void Connect(CServer const& server, Credentials const& credentials) override;
 	virtual void List(CServerPath const& path = CServerPath(), std::wstring const& subDir = std::wstring(), int flags = 0) override;
 	void ChangeDir(CServerPath const& path = CServerPath(), std::wstring const& subDir = std::wstring(), bool link_discovery = false);
-	virtual void FileTransfer(std::wstring const& localFile, CServerPath const& remotePath,
-		std::wstring const& remoteFile, transfer_flags const& flags) override;
+	virtual void FileTransfer(CFileTransferCommand const& cmd) override;
 	virtual void Delete(CServerPath const& path, std::vector<std::wstring>&& files) override;
 	virtual void RemoveDir(CServerPath const& path = CServerPath(), std::wstring const& subDir = std::wstring()) override;
 	virtual void Mkdir(CServerPath const& path) override;
@@ -31,8 +27,6 @@ public:
 	virtual void Chmod(CChmodCommand const& command) override;
 	virtual void Cancel() override;
 
-	virtual bool Connected() const override { return input_thread_.operator bool(); }
-
 	virtual bool SetAsyncRequestReply(CAsyncRequestNotification *pNotification) override;
 
 protected:
@@ -47,19 +41,23 @@ protected:
 	void ProcessReply(int result, std::wstring const& reply);
 
 	int SendCommand(std::wstring const& cmd, std::wstring const& show = std::wstring());
-	int AddToStream(std::wstring const& cmd);
-	int AddToStream(std::string const& cmd);
+	int AddToSendBuffer(std::wstring const& cmd);
+	int AddToSendBuffer(std::string const& cmd);
+	int SendToProcess();
 
 	virtual void wakeup(fz::direction::type const d) override;
 	void OnQuotaRequest(fz::direction::type const d);
 
+#ifndef FZ_WINDOWS
+	int shm_fd_{-1};
+#endif
 	std::unique_ptr<fz::process> process_;
-	std::unique_ptr<CSftpInputThread> input_thread_;
+	std::unique_ptr<SftpInputParser> input_parser_;
 
 	virtual void operator()(fz::event_base const& ev) override;
 	void OnSftpEvent(sftp_message const& message);
+	void OnProcessEvent(fz::process* p, fz::process_event_flag const& f);
 	void OnSftpListEvent(sftp_list_message const& message);
-	void OnTerminate(std::wstring const& error);
 
 	std::wstring m_requestPreamble;
 	std::wstring m_requestInstruction;
@@ -69,6 +67,8 @@ protected:
 	int result_{};
 	std::wstring response_;
 
+	fz::buffer send_buffer_;
+
 	friend class CProtocolOpData<CSftpControlSocket>;
 	friend class CSftpChangeDirOpData;
 	friend class CSftpChmodOpData;
