@@ -24,6 +24,10 @@
 
 #include <assert.h>
 
+#ifndef FZ_WINDOWS
+#include <unistd.h>
+#endif
+
 CStorjControlSocket::CStorjControlSocket(CFileZillaEnginePrivate & engine)
 	: CControlSocket(engine)
 {
@@ -49,10 +53,9 @@ void CStorjControlSocket::List(CServerPath const& path, std::wstring const& subD
 	Push(std::make_unique<CStorjListOpData>(*this, path, subDir, flags));
 }
 
-void CStorjControlSocket::FileTransfer(std::wstring const& localFile, CServerPath const& remotePath,
-						 std::wstring const& remoteFile, transfer_flags const& flags)
+void CStorjControlSocket::FileTransfer(CFileTransferCommand const& cmd)
 {
-	auto pData = std::make_unique<CStorjFileTransferOpData>(*this, localFile, remoteFile, remotePath, flags);
+	auto pData = std::make_unique<CStorjFileTransferOpData>(*this, cmd);
 	Push(std::move(pData));
 }
 
@@ -121,12 +124,14 @@ void CStorjControlSocket::OnStorjEvent(storj_message const& message)
 	case storjEvent::Status:
 		log_raw(logmsg::status, message.text[0]);
 		break;
+/* Unused
 	case storjEvent::Recv:
-		SetActive(CFileZillaEngine::recv);
+		RecordActivity(activity_logger::recv, fz::to_integral<uint64_t>(message.text[0]);
 		break;
 	case storjEvent::Send:
-		SetActive(CFileZillaEngine::send);
+		RecordActivity(activity_logger::send, fz::to_integral<uint64_t>(message.text[0]);
 		break;
+*/
 	case storjEvent::Listentry:
 		if (operations_.empty() || operations_.back()->opId != Command::list) {
 			log(logmsg::debug_warning, L"storjEvent::Listentry outside list operation, ignoring.");
@@ -146,9 +151,9 @@ void CStorjControlSocket::OnStorjEvent(storj_message const& message)
 			if (!operations_.empty() && operations_.back()->opId == Command::transfer) {
 				auto & data = static_cast<CStorjFileTransferOpData &>(*operations_.back());
 
-				SetActive(data.download() ? CFileZillaEngine::recv : CFileZillaEngine::send);
+				RecordActivity(data.download() ? activity_logger::recv : activity_logger::send, value);
 
-				bool tmp;
+				bool tmp{};
 				CTransferStatus status = engine_.transfer_status_.Get(tmp);
 				if (!status.empty() && !status.madeProgress) {
 					if (data.download()) {
@@ -163,10 +168,25 @@ void CStorjControlSocket::OnStorjEvent(storj_message const& message)
 					}
 				}
 			}
+			else {
+				SetAlive();
+			}
 
 			engine_.transfer_status_.Update(value);
 		}
 		break;
+	case storjEvent::io_nextbuf:
+		if (!operations_.empty() && operations_.back()->opId == Command::transfer) {
+			auto & data = static_cast<CStorjFileTransferOpData&>(*operations_.back());
+			data.OnNextBufferRequested(fz::to_integral<uint64_t>(message.text[0]));
+		}
+		break;
+	case storjEvent::io_finalize:
+		if (!operations_.empty() && operations_.back()->opId == Command::transfer) {
+			auto & data = static_cast<CStorjFileTransferOpData&>(*operations_.back());
+			data.OnFinalizeRequested(fz::to_integral<uint64_t>(message.text[0]));
+		}
+		break;
 	default:
 		log(logmsg::debug_warning, L"Message type %d not handled", message.type);
 		break;
@@ -208,19 +228,28 @@ int CStorjControlSocket::SendCommand(std::wstring const& cmd, std::wstring const
 
 int CStorjControlSocket::AddToStream(std::wstring const& cmd)
 {
-	if (!process_) {
-		ResetOperation(FZ_REPLY_INTERNALERROR);
-		return false;
-	}
-
 	std::string const str = ConvToServer(cmd, true);
 	if (str.empty()) {
 		log(logmsg::error, _("Could not convert command to server encoding"));
 		return FZ_REPLY_ERROR;
 	}
 
-	if (!process_->write(str)) {
-		return FZ_REPLY_ERROR | FZ_REPLY_DISCONNECTED;
+	return AddToStream(str);
+}
+
+int CStorjControlSocket::AddToStream(std::string_view cmd)
+{
+	if (!process_) {
+		ResetOperation(FZ_REPLY_INTERNALERROR);
+		return false;
+	}
+
+	while (!cmd.empty()) {
+		fz::rwresult written = process_->write(cmd);
+		if (!written) {
+			return FZ_REPLY_ERROR | FZ_REPLY_DISCONNECTED;
+		}
+		cmd = cmd.substr(written.value_);
 	}
 
 	return FZ_REPLY_WOULDBLOCK;
@@ -320,6 +349,14 @@ int CStorjControlSocket::DoClose(int nErrorCode)
 		event_loop_.filter_events(threadEventsFilter);
 	}
 	process_.reset();
+
+#ifndef FZ_WINDOWS
+	if (shm_fd_ != -1) {
+		close(shm_fd_);
+		shm_fd_ = -1;
+	}
+#endif
+
 	return CControlSocket::DoClose(nErrorCode);
 }
 
