@@ -26,33 +26,32 @@
 #include <libfilezilla/time.hpp>
 #include <libfilezilla/tls_info.hpp>
 
-enum NotificationId
+enum NotificationId : unsigned int
 {
 	nId_logmsg,				// notification about new messages for the message log
 	nId_operation,			// operation reply codes
-	nId_connection,			// connection information: connects, disconnects, timeouts etc..
 	nId_transferstatus,		// transfer information: bytes transferred, transfer speed and such
 	nId_listing,			// directory listings
 	nId_asyncrequest,		// asynchronous request
-	nId_active,				// sent if data gets either received or sent
-	nId_data,				// for memory downloads, indicates that new data is available.
 	nId_sftp_encryption,	// information about key exchange, encryption algorithms and so on for SFTP
 	nId_local_dir_created,	// local directory has been created
-	nId_serverchange		// With some protocols, actual server identity isn't known until after logon
+	nId_serverchange,		// With some protocols, actual server identity isn't known until after logon
+	nId_ftp_tls_resumption
 };
 
 // Async request IDs
-enum RequestId
+enum RequestId : unsigned int
 {
-	reqId_fileexists,         // Target file already exists, awaiting further instructions
-	reqId_interactiveLogin,   // gives a challenge prompt for a password
-	reqId_hostkey,            // used only by SSH/SFTP to indicate new host key
-	reqId_hostkeyChanged,     // used only by SSH/SFTP to indicate changed host key
-	reqId_certificate,        // sent after a successful TLS handshake to allow certificate
-	                          // validation.
-	reqId_insecure_connection // If using opportunistic FTP over TLS, or a completely
-	                          // unprotected protocol ask user whether he really wants
-	                          // to use a plaintext connection.
+	reqId_fileexists,          // Target file already exists, awaiting further instructions
+	reqId_interactiveLogin,    // gives a challenge prompt for a password
+	reqId_hostkey,             // used only by SSH/SFTP to indicate new host key
+	reqId_hostkeyChanged,      // used only by SSH/SFTP to indicate changed host key
+	reqId_certificate,         // sent after a successful TLS handshake to allow certificate
+	                           // validation.
+	reqId_insecure_connection, // If using opportunistic FTP over TLS, or a completely
+	                           // unprotected protocol ask user whether he really wants
+	                           // to use a plaintext connection.
+	reqId_tls_no_resumption
 };
 
 class FZC_PUBLIC_SYMBOL CNotification
@@ -186,9 +185,12 @@ public:
 	// Set overwriteAction to the desired action
 	OverwriteAction overwriteAction{unknown};
 
-	// Set to new filename if overwriteAction is rename. Might trigger further
+	// On uploads: Set to new filename if overwriteAction is rename. Might trigger further
 	// file exists notifications if new target file exists as well.
 	std::wstring newName;
+
+	// On downloads: New writer if overwriteAction is rename
+	writer_factory_holder new_writer_factory_;
 };
 
 class FZC_PUBLIC_SYMBOL CInteractiveLoginNotification final : public CAsyncRequestNotification
@@ -196,7 +198,8 @@ class FZC_PUBLIC_SYMBOL CInteractiveLoginNotification final : public CAsyncReque
 public:
 	enum type {
 		interactive,
-		keyfile
+		keyfile,
+		totp
 	};
 
 	CInteractiveLoginNotification(type t, std::wstring const& challenge, bool repeated);
@@ -224,17 +227,6 @@ protected:
 	bool const m_repeated;
 };
 
-// Indicate network action.
-class FZC_PUBLIC_SYMBOL CActiveNotification final : public CNotificationHelper<nId_active>
-{
-public:
-	explicit CActiveNotification(int direction);
-
-	int GetDirection() const { return m_direction; }
-protected:
-	const int m_direction;
-};
-
 class FZC_PUBLIC_SYMBOL CTransferStatus final
 {
 public:
@@ -283,8 +275,7 @@ public:
 	virtual ~CSftpEncryptionDetails() = default;
 
 	std::wstring hostKeyAlgorithm;
-	std::wstring hostKeyFingerprintMD5;
-	std::wstring hostKeyFingerprintSHA256;
+	std::wstring hostKeyFingerprint;
 	std::wstring kexAlgorithm;
 	std::wstring kexHash;
 	std::wstring kexCurve;
@@ -320,25 +311,6 @@ protected:
 	const bool m_changed;
 };
 
-class FZC_PUBLIC_SYMBOL CDataNotification final : public CNotificationHelper<nId_data>
-{
-public:
-	CDataNotification(char* pData, size_t len);
-	virtual ~CDataNotification();
-
-	CDataNotification(CDataNotification const&) = delete;
-	CDataNotification& operator=(CDataNotification const&) = delete;
-
-	char* Detach(size_t& len);
-
-	char const* data() const { return m_pData; }
-	size_t size() const { return m_len; }
-
-protected:
-	char* m_pData;
-	size_t m_len;
-};
-
 class FZC_PUBLIC_SYMBOL CCertificateNotification final : public CAsyncRequestNotification
 {
 public:
@@ -379,7 +351,32 @@ public:
 	    : newServer_(server)
 	{}
 
-	CServer newServer_;
+	CServer newServer_{};
+};
+
+class FZC_PUBLIC_SYMBOL FtpTlsResumptionNotification final : public CNotificationHelper<nId_ftp_tls_resumption>
+{
+public:
+	FtpTlsResumptionNotification() = default;
+
+	explicit FtpTlsResumptionNotification(CServer const& server)
+	    : server_(server)
+	{}
+
+	CServer const server_{};
+};
+
+class FZC_PUBLIC_SYMBOL FtpTlsNoResumptionNotification final : public CAsyncRequestNotification
+{
+public:
+	FtpTlsNoResumptionNotification(CServer const& server)
+	    : server_(server)
+	{}
+
+	virtual RequestId GetRequestID() const { return reqId_tls_no_resumption; }
+
+	CServer const server_;
+	bool allow_{};
 };
 
 #endif
