@@ -12,10 +12,11 @@
 #include "msgbox.h"
 #include "themeprovider.h"
 #include "wxfilesystem_blob_handler.h"
+#include "renderer.h"
+#include "../commonui/fz_paths.h"
 #include "../include/version.h"
 #include <libfilezilla/local_filesys.hpp>
 #include <libfilezilla/translate.hpp>
-
 #include <wx/evtloop.h>
 
 #ifdef WITH_LIBDBUS
@@ -44,69 +45,6 @@ IMPLEMENT_APP_NO_MAIN(CFileZillaApp)
   #error Please build wxWidgets with support for positional arguments.
 #endif
 
-namespace {
-#if FZ_WINDOWS
-std::wstring const PATH_SEP = L";";
-#else
-std::wstring const PATH_SEP = L":";
-#endif
-}
-
-// If non-empty, always terminated by a separator
-std::wstring GetOwnExecutableDir()
-{
-#ifdef FZ_WINDOWS
-	// Add executable path
-	std::wstring path;
-	path.resize(4095);
-	DWORD res;
-	while (true) {
-		res = GetModuleFileNameW(0, &path[0], path.size() - 1);
-		if (!res) {
-			// Failure
-			return std::wstring();
-		}
-
-		if (res < path.size() - 1) {
-			path.resize(res);
-			break;
-		}
-
-		path.resize(path.size() * 2 + 1);
-	}
-	size_t pos = path.find_last_of(L"\\/");
-	if (pos != std::wstring::npos) {
-		return path.substr(0, pos + 1);
-	}
-#elif defined(FZ_MAC)
-	// Fixme: Remove wxDependency and move entire function to libfilezilla
-	std::wstring executable = wxStandardPaths::Get().GetExecutablePath().ToStdWstring();
-	size_t pos = executable.rfind('/');
-	if (pos != std::wstring::npos) {
-		return executable.substr(0, pos + 1);
-	}
-#else
-	std::string path;
-	path.resize(4095);
-	while (true) {
-		int res = readlink("/proc/self/exe", &path[0], path.size());
-		if (res < 0) {
-			return std::wstring();
-		}
-		if (static_cast<size_t>(res) < path.size()) {
-			path.resize(res);
-			break;
-		}
-		path.resize(path.size() * 2 + 1);
-	}
-	size_t pos = path.rfind('/');
-	if (pos != std::wstring::npos) {
-		return fz::to_wstring(path.substr(0, pos + 1));
-	}
-#endif
-	return std::wstring();
-}
-
 CFileZillaApp::CFileZillaApp()
 {
 	m_profile_start = fz::monotonic_clock::now();
@@ -116,12 +54,11 @@ CFileZillaApp::CFileZillaApp()
 CFileZillaApp::~CFileZillaApp()
 {
 	themeProvider_.reset();
-	COptions::Destroy();
 }
 
 void CFileZillaApp::InitLocale()
 {
-	wxString language = COptions::Get()->get_string(OPTION_LANGUAGE);
+	wxString language = options_->get_string(OPTION_LANGUAGE);
 	const wxLanguageInfo* pInfo = wxLocale::FindLanguageInfo(language);
 	if (!language.empty()) {
 #ifdef __WXGTK__
@@ -159,18 +96,18 @@ void CFileZillaApp::InitLocale()
 			error += _("Please make sure the requested locale is installed on your system.");
 			wxMessageBoxEx(error, _("Failed to change language"), wxICON_EXCLAMATION);
 
-			COptions::Get()->set(OPTION_LANGUAGE, _T(""));
+			options_->set(OPTION_LANGUAGE, _T(""));
 		}
 #else
 		if (!pInfo || !SetLocale(pInfo->Language)) {
 			for (language = GetFallbackLocale(language); !language.empty(); language = GetFallbackLocale(language)) {
 				const wxLanguageInfo* fallbackInfo = wxLocale::FindLanguageInfo(language);
 				if (fallbackInfo && SetLocale(fallbackInfo->Language)) {
-					COptions::Get()->set(OPTION_LANGUAGE, language.ToStdWstring());
+					options_->set(OPTION_LANGUAGE, language.ToStdWstring());
 					return;
 				}
 			}
-			COptions::Get()->set(OPTION_LANGUAGE, std::wstring());
+			options_->set(OPTION_LANGUAGE, std::wstring());
 			if (pInfo && !pInfo->Description.empty()) {
 				wxMessageBoxEx(wxString::Format(_("Failed to set language to %s (%s), using default system language"), pInfo->Description, language), _("Failed to change language"), wxICON_EXCLAMATION);
 			}
@@ -202,6 +139,8 @@ bool CFileZillaApp::OnInit()
 {
 	AddStartupProfileRecord("CFileZillaApp::OnInit()");
 
+	SetAppDisplayName("FileZilla");
+
 	// Turn off idle events, we don't need them
 	wxIdleEvent::SetMode(wxIDLE_PROCESS_SPECIFIED);
 
@@ -229,6 +168,8 @@ bool CFileZillaApp::OnInit()
 	wxSystemOptions::SetOption(_T("mac.textcontrol-use-spell-checker"), 0);
 #endif
 
+	InitRenderer();
+
 	int cmdline_result = ProcessCommandLine();
 	if (!cmdline_result) {
 		return false;
@@ -253,9 +194,8 @@ bool CFileZillaApp::OnInit()
 		wxSetEnv("PUTTYDIR", home + L".config/putty");
 	}
 #endif
-	InitDefaultsDir();
 
-	COptions::Init();
+	options_ = std::make_unique<COptions>();
 
 	InitLocale();
 
@@ -273,18 +213,17 @@ USE AT OWN RISK"), _T("Important Information"));
 	else {
 		std::wstring v = GetEnv("FZDEBUG");
 		if (v != L"1") {
-			COptions::Get()->set(OPTION_LOGGING_DEBUGLEVEL, 0);
-			COptions::Get()->set(OPTION_LOGGING_RAWLISTING, 0);
+			options_->set(OPTION_LOGGING_DEBUGLEVEL, 0);
+			options_->set(OPTION_LOGGING_RAWLISTING, 0);
 		}
 	}
 #endif
 
 	if (!LoadResourceFiles()) {
-		COptions::Destroy();
 		return false;
 	}
 
-	themeProvider_ = std::make_unique<CThemeProvider>();
+	themeProvider_ = std::make_unique<CThemeProvider>(*options_);
 	CheckExistsFzsftp();
 #if ENABLE_STORJ
 	CheckExistsFzstorj();
@@ -303,7 +242,7 @@ USE AT OWN RISK"), _T("Important Information"));
 	OSXSandboxUserdirs::Get().Load();
 
 	if (OSXSandboxUserdirs::Get().GetDirs().empty()) {
-		CWelcomeDialog welcome;
+		CWelcomeDialog welcome(*options_, nullptr);
 		welcome.Run(nullptr, false);
 		welcome_skip = true;
 
@@ -315,16 +254,16 @@ USE AT OWN RISK"), _T("Important Information"));
     }
 #endif
 
-	CMainFrame *frame = new CMainFrame();
+	CMainFrame *frame = new CMainFrame(*options_);
 	frame->Show(true);
 	SetTopWindow(frame);
 
 	if (!welcome_skip) {
-		CWelcomeDialog::RunDelayed(frame);
+		auto welcome = new CWelcomeDialog(*options_, frame);
+		welcome->RunDelayed();
 	}
 
 	frame->ProcessCommandLine();
-	frame->PostInitialize();
 
 	ShowStartupProfile();
 
@@ -335,7 +274,7 @@ int CFileZillaApp::OnExit()
 {
 	CContextManager::Get()->NotifyGlobalHandlers(STATECHANGE_QUITNOW);
 
-	COptions::Get()->Save();
+	options_->Save();
 
 #ifdef WITH_LIBDBUS
 	CSessionManager::Uninit();
@@ -352,106 +291,10 @@ int CFileZillaApp::OnExit()
 	return wxApp::OnExit();
 }
 
-bool CFileZillaApp::FileExists(std::wstring const& file) const
-{
-	return fz::local_filesys::get_file_type(fz::to_native(file), true) == fz::local_filesys::file;
-}
-
-CLocalPath CFileZillaApp::GetDataDir(std::vector<std::wstring> const& fileToFind, std::wstring const& prefixSub, bool searchSelfDir) const
-{
-	/*
-	 * Finding the resources in all cases is a difficult task,
-	 * due to the huge variety of diffent systems and their filesystem
-	 * structure.
-	 * Basically we just check a couple of paths for presence of the resources,
-	 * and hope we find them. If not, the user can still specify on the cmdline
-	 * and using environment variables where the resources are.
-	 *
-	 * At least on OS X it's simple: All inside application bundle.
-	 */
-
-	CLocalPath ret;
-
-	auto testPath = [&](std::wstring const& path) {
-		ret = CLocalPath(path);
-		if (ret.empty()) {
-			return false;
-		}
-
-		for (auto const& file : fileToFind) {
-			if (FileExists(ret.GetPath() + file)) {
-				return true;
-			}
-		}
-		return false;
-	};
-
-#ifdef __WXMAC__
-	(void)prefixSub;
-
-	if (searchSelfDir && testPath(wxStandardPaths::Get().GetDataDir().ToStdWstring())) {
-		return ret;
-	}
-
-	return CLocalPath();
-#else
-
-	// First try the user specified data dir.
-	if (searchSelfDir) {
-		if (testPath(GetEnv("FZ_DATADIR"))) {
-			return ret;
-		}
-	}
-
-	std::wstring selfDir = GetOwnExecutableDir();
-	if (!selfDir.empty()) {
-		if (searchSelfDir && testPath(selfDir)) {
-			return ret;
-		}
-
-		if (!prefixSub.empty() && selfDir.size() > 5 && fz::ends_with(selfDir, std::wstring(L"/bin/"))) {
-			std::wstring path = selfDir.substr(0, selfDir.size() - 4) + prefixSub + L"/";
-			if (testPath(path)) {
-				return ret;
-			}
-		}
-
-		// Development paths
-		if (searchSelfDir && selfDir.size() > 7 && fz::ends_with(selfDir, std::wstring(L"/.libs/"))) {
-			std::wstring path = selfDir.substr(0, selfDir.size() - 6);
-			if (FileExists(path + L"Makefile")) {
-				if (testPath(path)) {
-					return ret;
-				}
-			}
-		}
-	}
-
-	// Now scan through the path
-	if (!prefixSub.empty()) {
-		std::wstring path = GetEnv("PATH");
-		auto const segments = fz::strtok(path, PATH_SEP);
-
-		for (auto const& segment : segments) {
-			auto const cur = CLocalPath(segment).GetPath();
-			if (cur.size() > 5 && fz::ends_with(cur, std::wstring(L"/bin/"))) {
-				std::wstring path = cur.substr(0, cur.size() - 4) + prefixSub + L"/";
-				if (testPath(path)) {
-					return ret;
-				}
-			}
-		}
-	}
-
-	ret.clear();
-	return ret;
-#endif
-}
-
 bool CFileZillaApp::LoadResourceFiles()
 {
 	AddStartupProfileRecord("CFileZillaApp::LoadResourceFiles");
-	m_resourceDir = GetDataDir({L"resources/defaultfilters.xml"}, L"share/filezilla");
+	m_resourceDir = GetFZDataDir({L"resources/defaultfilters.xml"}, L"share/filezilla");
 
 	wxImage::AddHandler(new wxPNGHandler());
 
@@ -470,38 +313,16 @@ bool CFileZillaApp::LoadResourceFiles()
 	return true;
 }
 
-bool CFileZillaApp::InitDefaultsDir()
-{
-	AddStartupProfileRecord("InitDefaultsDir");
-#ifdef __WXGTK__
-	m_defaultsDir = COptions::GetUnadjustedSettingsDir();
-	if (m_defaultsDir.empty() || !FileExists(m_defaultsDir.GetPath() + L"fzdefaults.xml")) {
-		if (FileExists(L"/etc/filezilla/fzdefaults.xml")) {
-			m_defaultsDir.SetPath(L"/etc/filezilla");
-		}
-		else {
-			m_defaultsDir.clear();
-		}
-	}
-
-#endif
-	if (m_defaultsDir.empty()) {
-		m_defaultsDir = GetDataDir({L"fzdefaults.xml"}, L"share/filezilla");
-	}
-
-	return !m_defaultsDir.empty();
-}
-
 bool CFileZillaApp::LoadLocales()
 {
 	AddStartupProfileRecord("CFileZillaApp::LoadLocales");
-	m_localesDir = GetDataDir({L"locales/de/filezilla.mo"}, std::wstring());
+	m_localesDir = GetFZDataDir({L"locales/de/filezilla.mo"}, std::wstring());
 	if (!m_localesDir.empty()) {
 		m_localesDir.AddSegment(_T("locales"));
 	}
 #ifndef __WXMAC__
 	else {
-		m_localesDir = GetDataDir({L"de/filezilla.mo", L"de/LC_MESSAGES/filezilla.mo"}, L"share/locale", false);
+		m_localesDir = GetFZDataDir({L"de/filezilla.mo", L"de/LC_MESSAGES/filezilla.mo"}, L"share/locale", false);
 	}
 #endif
 	if (!m_localesDir.empty()) {
@@ -608,90 +429,17 @@ void CFileZillaApp::CheckExistsFzstorj()
 
 void CFileZillaApp::CheckExistsTool(std::wstring const& tool, std::wstring const& buildRelPath, char const* env, engineOptions setting, std::wstring const& description)
 {
-	// Get the correct path to the specified tool
+	std::wstring const executable = FindTool(tool, buildRelPath, env);
 
-	bool found = false;
-	std::wstring executable;
-
-	std::wstring program = tool;
-#ifdef __WXMAC__
-	(void)buildRelPath;
-
-	// On Mac we only look inside the bundle
-	std::wstring path = GetOwnExecutableDir();
-	if (!path.empty()) {
-		executable = path + '/' + tool;
-		if (FileExists(executable)) {
-			found = true;
-		}
-	}
-#else
-
-#ifdef __WXMSW__
-	program += L".exe";
+	if (executable.empty()) {
+		std::wstring program = tool;
+#if FZ_WINDOWS
+		program += L".exe";
 #endif
-
-	// First check the given environment variable
-	executable = GetEnv(env);
-	if (!executable.empty()) {
-		if (FileExists(executable)) {
-			found = true;
-		}
-	}
-
-	if (!found) {
-		std::wstring path = GetOwnExecutableDir();
-		if (!path.empty()) {
-			executable = path + program;
-			if (FileExists(executable)) {
-				found = true;
-			}
-			else {
-				// Check if running from build dir
-				if (path.size() > 7 && fz::ends_with(path, std::wstring(L"/.libs/"))) {
-					if (FileExists(path.substr(0, path.size() - 6) + L"Makefile")) {
-						executable = path + L"../" + buildRelPath + program;
-						if (FileExists(executable)) {
-							found = true;
-						}
-					}
-				}
-				else if (FileExists(path + L"Makefile")) {
-					executable = path + buildRelPath + program;
-					if (FileExists(executable)) {
-						found = true;
-					}
-				}
-			}
-		}
-	}
-
-	if (!found) {
-		// Check PATH
-		std::wstring path = GetEnv("PATH");
-		auto const segments = fz::strtok(path, PATH_SEP);
-		for (auto const& segment : segments) {
-			auto const cur = CLocalPath(segment).GetPath();
-			executable = cur + program;
-			if (!cur.empty() && FileExists(executable)) {
-				found = true;
-				break;
-			}
-		}
-	}
-#endif
-
-	if (!found) {
-		// Quote path if it contains spaces
-		if (executable.find(' ') != std::wstring::npos && executable.front() != '"' && executable.front() != '\'') {
-			executable = L"\"" + executable + L"\"";
-		}
-
 		wxMessageBoxEx(fz::sprintf(fztranslate("%s could not be found. Without this component of FileZilla, %s will not work.\n\nPossible solutions:\n- Make sure %s is in a directory listed in your PATH environment variable.\n- Set the full path to %s in the %s environment variable."), program, description, program, program, env),
 			_("File not found"), wxICON_ERROR | wxOK);
-		executable.clear();
 	}
-	COptions::Get()->set(setting, executable);
+	options_->set(setting, executable);
 }
 
 #ifdef __WXMSW__
@@ -751,7 +499,7 @@ void CFileZillaApp::ShowStartupProfile()
 		wxString msg = _T("Profile:\n");
 
 		size_t const max_digits = fz::to_string((m_startupProfile.back().first - m_profile_start).get_milliseconds()).size();
-		
+
 		int64_t prev{};
 		for (auto const& p : m_startupProfile) {
 			auto const diff = p.first - m_profile_start;
@@ -783,7 +531,7 @@ void CFileZillaApp::ShowStartupProfile()
 
 std::wstring CFileZillaApp::GetSettingsFile(std::wstring const& name) const
 {
-	return COptions::Get()->get_string(OPTION_DEFAULT_SETTINGSDIR) + name + _T(".xml");
+	return options_->get_string(OPTION_DEFAULT_SETTINGSDIR) + name + _T(".xml");
 }
 
 #if defined(__MINGW32__)
