@@ -4,7 +4,7 @@
 #include "dragdropmanager.h"
 #include "drop_target_ex.h"
 #include "filezillaapp.h"
-#include "filter.h"
+#include "filter_manager.h"
 #include "file_utils.h"
 #include "graphics.h"
 #include "inputdialog.h"
@@ -18,7 +18,6 @@
 #include <wx/menu.h>
 
 #ifdef __WXMSW__
-#include <wx/msw/registry.h>
 #include <shlobj.h>
 #include <dbt.h>
 #include "volume_enumerator.h"
@@ -46,15 +45,6 @@ public:
 	{
 	}
 
-	void ClearDropHighlight()
-	{
-		const wxTreeItemId dropHighlight = m_pLocalTreeView->m_dropHighlight;
-		if (dropHighlight != wxTreeItemId()) {
-			m_pLocalTreeView->SetItemDropHighlight(dropHighlight, false);
-			m_pLocalTreeView->m_dropHighlight = wxTreeItemId();
-		}
-	}
-
 	std::wstring GetDirFromItem(wxTreeItemId const& item)
 	{
 		std::wstring const dir = m_pLocalTreeView->GetDirFromItem(item);
@@ -111,26 +101,27 @@ public:
 			pDragDropManager->pDropTarget = m_pLocalTreeView;
 		}
 
-		auto const format = m_pDataObject->GetReceivedFormat();
+		auto const format = GetReceivedFormat();
 		if (format == m_pFileDataObject->GetFormat()) {
 			m_pLocalTreeView->m_state.HandleDroppedFiles(m_pFileDataObject, path, def == wxDragCopy);
 		}
-		else if (format == m_pLocalDataObject->GetFormat()) {
-			m_pLocalTreeView->m_state.HandleDroppedFiles(m_pLocalDataObject, path, def == wxDragCopy);
+		else if (format == LocalDataObjectFormat()) {
+			m_pLocalTreeView->m_state.HandleDroppedFiles(GetLocalDataObject(), path, def == wxDragCopy);
 		}
 		else {
-			if (m_pRemoteDataObject->GetProcessId() != (int)wxGetProcessId()) {
+			auto * obj = GetRemoteDataObject();
+			if (obj->GetProcessId() != (int)wxGetProcessId()) {
 				wxMessageBoxEx(_("Drag&drop between different instances of FileZilla has not been implemented yet."));
 				return wxDragNone;
 			}
 
 			auto & state = m_pLocalTreeView->m_state;
-			if (!state.GetSite() || m_pRemoteDataObject->GetSite().server != state.GetSite().server) {
+			if (!state.GetSite() || obj->GetSite().server != state.GetSite().server) {
 				wxMessageBoxEx(_("Drag&drop between different servers has not been implemented yet."));
 				return wxDragNone;
 			}
 
-			if (!state.DownloadDroppedFiles(m_pRemoteDataObject, path)) {
+			if (!state.DownloadDroppedFiles(obj, path)) {
 				return wxDragNone;
 			}
 		}
@@ -140,11 +131,11 @@ public:
 
 	virtual bool OnDrop(wxCoord x, wxCoord y)
 	{
-		if (!CScrollableDropTarget<wxTreeCtrlEx>::OnDrop(x, y)) {
+		if (!CFileDropTarget::OnDrop(x, y)) {
 			return false;
 		}
 
-		ClearDropHighlight();
+		m_pLocalTreeView->ClearDropHighlight();
 
 		wxTreeItemId hit = GetHit(wxPoint(x, y));
 		if (!hit) {
@@ -159,41 +150,36 @@ public:
 		return true;
 	}
 
-	wxTreeItemId DisplayDropHighlight(wxPoint point)
+	wxTreeItemId DisplayDropHighlight(wxPoint const& point) override
 	{
 		wxTreeItemId hit = GetHit(point);
 		if (!hit) {
-			ClearDropHighlight();
+			m_pLocalTreeView->ClearDropHighlight();
 			return hit;
 		}
 
 		std::wstring dir = GetDirFromItem(hit);
 
 		if (dir.empty()) {
-			ClearDropHighlight();
+			m_pLocalTreeView->ClearDropHighlight();
 			return wxTreeItemId();
 		}
-
-		const wxTreeItemId dropHighlight = m_pLocalTreeView->m_dropHighlight;
-		if (dropHighlight != wxTreeItemId()) {
-			m_pLocalTreeView->SetItemDropHighlight(dropHighlight, false);
+		else {
+			m_pLocalTreeView->DisplayDropHighlight(hit);
 		}
 
-		m_pLocalTreeView->SetItemDropHighlight(hit, true);
-		m_pLocalTreeView->m_dropHighlight = hit;
-
 		return hit;
 	}
 
 	virtual wxDragResult OnDragOver(wxCoord x, wxCoord y, wxDragResult def)
 	{
-		def = CScrollableDropTarget<wxTreeCtrlEx>::OnDragOver(x, y, def);
+		def = CScrollableDropTarget::OnDragOver(x, y, def);
 
 		if (def == wxDragError ||
 			def == wxDragNone ||
 			def == wxDragCancel)
 		{
-			ClearDropHighlight();
+			m_pLocalTreeView->ClearDropHighlight();
 			return def;
 		}
 
@@ -211,13 +197,13 @@ public:
 
 	virtual void OnLeave()
 	{
-		CScrollableDropTarget<wxTreeCtrlEx>::OnLeave();
-		ClearDropHighlight();
+		CScrollableDropTarget::OnLeave();
+		m_pLocalTreeView->ClearDropHighlight();
 	}
 
 	virtual wxDragResult OnEnter(wxCoord x, wxCoord y, wxDragResult def)
 	{
-		def = CScrollableDropTarget<wxTreeCtrlEx>::OnEnter(x, y, def);
+		def = CScrollableDropTarget::OnEnter(x, y, def);
 		return OnDragOver(x, y, def);
 	}
 
@@ -641,26 +627,28 @@ std::wstring CLocalTreeView::GetDirFromItem(wxTreeItemId item)
 	while (item) {
 #ifdef __WXMSW__
 		if (item == m_desktop) {
-			wxChar path[MAX_PATH + 1];
-			if (SHGetFolderPath(0, CSIDL_DESKTOPDIRECTORY, 0, SHGFP_TYPE_CURRENT, path) != S_OK) {
-				if (SHGetFolderPath(0, CSIDL_DESKTOP, 0, SHGFP_TYPE_CURRENT, path) != S_OK) {
-					wxMessageBoxEx(_("Failed to get desktop path"));
-					return L"/";
-				}
+			wchar_t* out{};
+			if (SHGetKnownFolderPath(FOLDERID_Desktop, 0, 0, &out) != S_OK) {
+				wxMessageBoxEx(_("Failed to get desktop path"));
+				return L"/";
 			}
-			dir = path;
+			dir = out;
+			CoTaskMemFree(out);
+
 			if (dir.empty() || dir.back() != separator) {
 				dir += separator;
 			}
 			return dir;
 		}
 		else if (item == m_documents) {
-			wxChar path[MAX_PATH + 1];
-			if (SHGetFolderPath(0, CSIDL_PERSONAL, 0, SHGFP_TYPE_CURRENT, path) != S_OK) {
+			wchar_t* out{};
+			if (SHGetKnownFolderPath(FOLDERID_Documents, 0, 0, &out) != S_OK) {
 				wxMessageBoxEx(_("Failed to get 'My Documents' path"));
 				return L"/";
 			}
-			dir = path;
+			dir = out;
+			CoTaskMemFree(out);
+
 			if (dir.empty() || dir.back() != separator) {
 				dir += separator;
 			}
@@ -697,18 +685,18 @@ std::wstring CLocalTreeView::GetDirFromItem(wxTreeItemId item)
 
 void CLocalTreeView::UpdateSortMode()
 {
-	CFileListCtrlSortBase::NameSortMode sortMode;
+	NameSortMode sortMode;
 	switch (COptions::Get()->get_int(OPTION_FILELIST_NAMESORT))
 	{
 	case 0:
 	default:
-		sortMode = CFileListCtrlSortBase::namesort_caseinsensitive;
+		sortMode = NameSortMode::case_insensitive;
 		break;
 	case 1:
-		sortMode = CFileListCtrlSortBase::namesort_casesensitive;
+		sortMode = NameSortMode::case_sensitive;
 		break;
 	case 2:
-		sortMode = CFileListCtrlSortBase::namesort_natural;
+		sortMode = NameSortMode::natural;
 		break;
 	}
 	sortFunction_ = CFileListCtrlSortBase::GetCmpFunction(sortMode);
@@ -924,7 +912,7 @@ void CLocalTreeView::OnStateChange(t_statechange_notifications notification, std
 		SetDir(m_state.GetLocalDir().GetPath());
 	}
 	else if (notification == STATECHANGE_SERVER) {
-		m_windowTinter->SetBackgroundTint(m_state.GetSite().m_colour);
+		m_windowTinter->SetBackgroundTint(site_colour_to_wx(m_state.GetSite().m_colour));
 	}
 	else {
 		wxASSERT(notification == STATECHANGE_APPLYFILTER);
@@ -969,21 +957,22 @@ void CLocalTreeView::OnBeginDrag(wxTreeEvent& event)
 	}
 #endif
 
-	CDragDropManager* pDragDropManager = CDragDropManager::Init();
-	pDragDropManager->pDragSource = this;
-
 #ifdef __WXMAC__
 	// Don't use wxFileDataObject on Mac, crashes on Mojave, wx bug #18232
 	CLocalDataObject obj;
-	obj.AddFile(dir);
 #else
 	wxFileDataObject obj;
-	obj.AddFile(dir);
 #endif
 
-	wxDropSource source(this);
+	CDragDropManager* pDragDropManager = CDragDropManager::Init();
+	pDragDropManager->pDragSource = this;
+
+	obj.AddFile(dir);
+	pDragDropManager->dragDataObject = &obj;
+
+	DropSource source(this);
 	source.SetData(obj);
-	int res = source.DoDragDrop(wxDrag_AllowMove);
+	int res = source.DoFileDragDrop(wxDrag_AllowMove);
 
 	bool handled_internally = pDragDropManager->pDropTarget != 0;
 
@@ -994,6 +983,13 @@ void CLocalTreeView::OnBeginDrag(wxTreeEvent& event)
 		// externally, the internal handlers do this for us already
 		m_state.RefreshLocal();
 	}
+
+#ifdef __WXMAC__
+	if (!source.m_OutDir.IsEmpty()) {
+		CLocalPath target(source.m_OutDir.ToStdWstring());
+		m_state.HandleDroppedFiles(&obj, target, true);
+	}
+#endif
 }
 
 #ifdef __WXMSW__
@@ -1184,7 +1180,7 @@ void CLocalTreeView::OnMenuUpload(wxCommandEvent& event)
 	bool const queue_only = event.GetId() == XRCID("ID_ADDTOQUEUE");
 
 	CFilterManager filter;
-	recursiveOperation->StartRecursiveOperation(CRecursiveOperation::recursive_transfer, filter.GetActiveFilters(), !queue_only);
+	recursiveOperation->StartRecursiveOperation(recursive_operation::recursive_transfer, filter.GetActiveFilters(), !queue_only);
 }
 
 // Create a new Directory
@@ -1570,7 +1566,7 @@ wxTreeItemId CLocalTreeView::AddDrive(wxChar letter)
 	drive += _T(":");
 
 	long drivesToHide = CVolumeDescriptionEnumeratorThread::GetDrivesToHide();
-	if( CVolumeDescriptionEnumeratorThread::IsHidden(drive.wc_str(), drivesToHide) ) {
+	if ( CVolumeDescriptionEnumeratorThread::IsHidden(drive.wc_str(), drivesToHide) ) {
 		return wxTreeItemId();
 	}
 
