@@ -1,35 +1,21 @@
 #ifndef FILEZILLA_INTERFACE_OPTIONS_HEADER
 #define FILEZILLA_INTERFACE_OPTIONS_HEADER
 
-#include "../include/local_path.h"
 #include "../include/engine_options.h"
 
-#include <libfilezilla/mutex.hpp>
+#include "../commonui/options.h"
 
 #include <wx/timer.h>
 
-enum interfaceOptions
+enum interfaceOptions : unsigned int
 {
 	// Default/internal options
-	OPTION_DEFAULT_SETTINGSDIR, // guaranteed to be (back)slash-terminated
-	OPTION_DEFAULT_KIOSKMODE,
-	OPTION_DEFAULT_DISABLEUPDATECHECK,
 	OPTION_DEFAULT_CACHE_DIR,
 
 	OPTION_NUMTRANSFERS,
-	OPTION_ASCIIBINARY,
-	OPTION_ASCIIFILES,
-	OPTION_ASCIINOEXT,
-	OPTION_ASCIIDOTFILE,
 	OPTION_LANGUAGE,
 	OPTION_CONCURRENTDOWNLOADLIMIT,
 	OPTION_CONCURRENTUPLOADLIMIT,
-	OPTION_UPDATECHECK,
-	OPTION_UPDATECHECK_INTERVAL,
-	OPTION_UPDATECHECK_LASTDATE,
-	OPTION_UPDATECHECK_LASTVERSION,
-	OPTION_UPDATECHECK_NEWVERSION,
-	OPTION_UPDATECHECK_CHECKBETA,
 	OPTION_DEBUG_MENU,
 	OPTION_FILEEXISTS_DOWNLOAD,
 	OPTION_FILEEXISTS_UPLOAD,
@@ -59,7 +45,6 @@ enum interfaceOptions
 	OPTION_EDIT_ALWAYSDEFAULT,
 	OPTION_EDIT_CUSTOMASSOCIATIONS,
 	OPTION_COMPARISONMODE,
-	OPTION_COMPARISON_THRESHOLD,
 	OPTION_SITEMANAGER_POSITION,
 	OPTION_ICONS_THEME,
 	OPTION_ICONS_SCALE,
@@ -99,55 +84,33 @@ enum interfaceOptions
 	OPTION_QUEUE_COMPLETION_COMMAND,
 	OPTION_DND_DISABLED,
 	OPTION_DISABLE_UPDATE_FOOTER,
-	OPTION_MASTERPASSWORDENCRYPTOR,
 	OPTION_TAB_DATA,
+	OPTION_SHOWN_OVERLAY,
 
 	// Has to be last element
 	OPTIONS_NUM
 };
 
-unsigned int register_interface_options();
-
 optionsIndex mapOption(interfaceOptions opt);
 
-std::wstring GetEnv(char const* name);
-
-class CXmlFile;
-class COptions final : public wxEvtHandler, public COptionsBase
+class COptions final : public wxEvtHandler, public XmlOptions
 {
 public:
-	static COptions* Get();
-	static void Init();
-	static void Destroy();
+	COptions();
+	virtual ~COptions();
+	
+	COptions(COptions const&) = delete;
+	COptions& operator=(COptions const&) = delete;
 
-	void Import(pugi::xml_node & element);
+	static COptions* Get();
+
+	CLocalPath GetCacheDirectory();
 
 	void Save(bool processChanged = true);
 
-	static CLocalPath GetUnadjustedSettingsDir();
-	CLocalPath GetCacheDirectory();
-
-	bool Cleanup(); // Removes all unknown elements from the XML
-
 protected:
-	COptions();
-	virtual ~COptions();
-
-	void Load(pugi::xml_node & settings, bool predefined, bool importing);
-
-	pugi::xml_node CreateSettingsXmlElement();
-
-	void LoadGlobalDefaultOptions();
-	CLocalPath InitSettingsDir();
-
-	virtual void process_changed(watched_options const& changed) override;
-	void set_xml_value(pugi::xml_node & settings, size_t opt, bool clean);
-
 	virtual void notify_changed() override;
-
-	bool dirty_{};
-	std::unique_ptr<CXmlFile> xmlFile_;
-
+	virtual void on_dirty() override;
 	static COptions* m_theOptions;
 
 	wxTimer m_save_timer;
