@@ -5,7 +5,9 @@
 #include "local_recursive_operation.h"
 #include "option_change_event_handler.h"
 #include "queue.h"
+#include "commandqueue.h"
 #include "queue_storage.h"
+#include "state.h"
 
 #include "../include/libfilezilla_engine.h"
 #include "../include/notification.h"
@@ -106,21 +108,24 @@ private:
 	bool trigger_{};
 };
 
-class CQueueView final : public CQueueViewBase, public COptionChangeEventHandler, public CGlobalStateEventHandler
+class CQueueView final : public CQueueViewBase,
+	public COptionChangeEventHandler, public CGlobalStateEventHandler, public CExclusiveHandler
 {
 	friend class CQueueViewDropTarget;
 	friend class CQueueViewFailed;
 	friend class CActionAfterBlocker;
 
 public:
-	CQueueView(CQueue* parent, int index, CMainFrame* pMainFrame, CAsyncRequestQueue* pAsyncRequestQueue);
+	CQueueView(CQueue* parent, int index, CMainFrame* pMainFrame, CAsyncRequestQueue* pAsyncRequestQueue, cert_store & certStore);
 	virtual ~CQueueView();
 
 	bool QueueFile(bool const queueOnly, bool const download,
 		std::wstring const& sourceFile, std::wstring const& targetFile,
 		CLocalPath const& localPath, CServerPath const& remotePath,
 		Site const& site, int64_t size, CEditHandler::fileType edit = CEditHandler::none,
-		QueuePriority priority = QueuePriority::normal);
+		QueuePriority priority = QueuePriority::normal, transfer_flags custom_flags = transfer_flags::none,
+		transfer_flags custom_flags_mask = transfer_flags::none,
+		std::wstring const& extraFlags = {});
 
 	void QueueFile_Finish(const bool start); // Need to be called after QueueFile
 	bool QueueFiles(const bool queueOnly, CLocalPath const& localPath, const CRemoteDataObject& dataObject);
@@ -139,27 +144,18 @@ public:
 	void RemoveAll();
 
 	void LoadQueue();
-	void LoadQueueFromXML();
 	void ImportQueue(pugi::xml_node element, bool updateSelections);
 
 	virtual void InsertItem(CServerItem* pServerItem, CQueueItem* pItem) override;
 
 	virtual void CommitChanges() override;
 
-	void ProcessNotification(CFileZillaEngine* pEngine, std::unique_ptr<CNotification>&& pNotification);
+	virtual void ProcessNotification(CFileZillaEngine* pEngine, std::unique_ptr<CNotification>&& pNotification) override;
 
-	void RenameFileInTransfer(CFileZillaEngine *pEngine, const wxString& newName, bool local);
+	void RenameFileInTransfer(CFileZillaEngine *pEngine, std::wstring const& newName, bool local, writer_factory_holder & new_writer);
 
 	static std::wstring ReplaceInvalidCharacters(std::wstring const& filename, bool includeQuotesAndBreaks = false);
 
-	// Get the current download speed as the sum of all active downloads.
-	// Unit is byte/s.
-	wxFileOffset GetCurrentDownloadSpeed();
-
-	// Get the current upload speed as the sum of all active uploads.
-	// Unit is byte/s.
-	wxFileOffset GetCurrentUploadSpeed();
-
 	std::shared_ptr<CActionAfterBlocker> GetActionAfterBlocker();
 
 protected:
@@ -259,6 +255,7 @@ protected:
 
 	CMainFrame* m_pMainFrame;
 	CAsyncRequestQueue* m_pAsyncRequestQueue;
+	cert_store & cert_store_;
 
 	std::list<CFileZillaEngine*> m_waitingForPassword;
 
@@ -282,10 +279,6 @@ protected:
 
 	CQueueStorage m_queue_storage;
 
-	// Get the current transfer speed.
-	// Unit is byte/s.
-	wxFileOffset GetCurrentSpeed(bool countDownload, bool countUpload);
-
 	void OnEngineEvent(CFileZillaEngine* engine);
 
 	void OnAskPassword();
@@ -308,7 +301,7 @@ protected:
 
 	void OnSetPriority(wxCommandEvent& event);
 
-	void OnExclusiveEngineRequestGranted(wxCommandEvent& event);
+	virtual void OnExclusiveEngineRequestGranted(unsigned int requestId) override;
 
 	void OnActionAfter(wxCommandEvent& event);
 	void OnActionAfterTimerTick();
