@@ -10,7 +10,7 @@
 #include "drop_target_ex.h"
 #include "edithandler.h"
 #include "filezillaapp.h"
-#include "filter.h"
+#include "filter_manager.h"
 #include "graphics.h"
 #include "infotext.h"
 #include "inputdialog.h"
@@ -20,6 +20,7 @@
 #include "remote_recursive_operation.h"
 #include "sizeformatting.h"
 #include "timeformatting.h"
+#include "../commonui/misc.h"
 
 #include <wx/clipbrd.h>
 #include <wx/menu.h>
@@ -85,8 +86,8 @@ public:
 			pDragDropManager->pDropTarget = m_pRemoteListView;
 		}
 
-		auto const format = m_pDataObject->GetReceivedFormat();
-		if (format == m_pFileDataObject->GetFormat() || format == m_pLocalDataObject->GetFormat()) {
+		auto const format = GetReceivedFormat();
+		if (format == m_pFileDataObject->GetFormat() || format == LocalDataObjectFormat()) {
 			std::wstring subdir;
 			int flags = 0;
 			int hit = m_pRemoteListView->HitTest(wxPoint(x, y), flags, 0);
@@ -101,24 +102,24 @@ public:
 					}
 				}
 			}
-
 			if (format == m_pFileDataObject->GetFormat()) {
 				m_pRemoteListView->m_state.UploadDroppedFiles(m_pFileDataObject, subdir, false);
 			}
 			else {
-				m_pRemoteListView->m_state.UploadDroppedFiles(m_pLocalDataObject, subdir, false);
+				m_pRemoteListView->m_state.UploadDroppedFiles(GetLocalDataObject(), subdir, false);
 			}
 			return wxDragCopy;
 		}
 
 		// At this point it's the remote data object.
+		auto * obj = GetRemoteDataObject();
 
-		if (m_pRemoteDataObject->GetProcessId() != (int)wxGetProcessId()) {
+		if (obj->GetProcessId() != (int)wxGetProcessId()) {
 			wxMessageBoxEx(_("Drag&drop between different instances of FileZilla has not been implemented yet."));
 			return wxDragNone;
 		}
 
-		if (m_pRemoteDataObject->GetSite().server != m_pRemoteListView->m_state.GetSite().server) {
+		if (obj->GetSite().server != m_pRemoteListView->m_state.GetSite().server) {
 			wxMessageBoxEx(_("Drag&drop between different servers has not been implemented yet."));
 			return wxDragNone;
 		}
@@ -151,15 +152,15 @@ public:
 		}
 
 		// Make sure target path is valid
-		if (target == m_pRemoteDataObject->GetServerPath()) {
+		if (target == obj->GetServerPath()) {
 			wxMessageBoxEx(_("Source and target of the drop operation are identical"));
 			return wxDragNone;
 		}
 
-		const std::vector<CRemoteDataObject::t_fileInfo>& files = m_pRemoteDataObject->GetFiles();
+		const std::vector<CRemoteDataObject::t_fileInfo>& files = obj->GetFiles();
 		for (auto const& info : files) {
 			if (info.dir) {
-				CServerPath dir = m_pRemoteDataObject->GetServerPath();
+				CServerPath dir = obj->GetServerPath();
 				dir.AddSegment(info.name);
 				if (dir == target) {
 					return wxDragNone;
@@ -173,7 +174,7 @@ public:
 
 		for (auto const& info : files) {
 			m_pRemoteListView->m_state.m_pCommandQueue->ProcessCommand(
-				new CRenameCommand(m_pRemoteDataObject->GetServerPath(), info.name, target, info.name)
+				new CRenameCommand(obj->GetServerPath(), info.name, target, info.name)
 				);
 		}
 
@@ -185,7 +186,7 @@ public:
 
 	virtual bool OnDrop(wxCoord x, wxCoord y)
 	{
-		CScrollableDropTarget<wxListCtrlEx>::OnDrop(x, y);
+		CScrollableDropTarget::OnDrop(x, y);
 		ClearDropHighlight();
 
 		if (!m_pRemoteListView->m_pDirectoryListing) {
@@ -195,7 +196,7 @@ public:
 		return true;
 	}
 
-	virtual int DisplayDropHighlight(wxPoint point)
+	virtual int DisplayDropHighlight(wxPoint const& point)
 	{
 		DoDisplayDropHighlight(point);
 		return -1;
@@ -244,7 +245,7 @@ public:
 
 	virtual wxDragResult OnDragOver(wxCoord x, wxCoord y, wxDragResult def)
 	{
-		def = CScrollableDropTarget<wxListCtrlEx>::OnDragOver(x, y, def);
+		def = CScrollableDropTarget::OnDragOver(x, y, def);
 
 		if (def == wxDragError ||
 			def == wxDragNone ||
@@ -280,13 +281,13 @@ public:
 
 	virtual void OnLeave()
 	{
-		CScrollableDropTarget<wxListCtrlEx>::OnLeave();
+		CScrollableDropTarget::OnLeave();
 		ClearDropHighlight();
 	}
 
 	virtual wxDragResult OnEnter(wxCoord x, wxCoord y, wxDragResult def)
 	{
-		def = CScrollableDropTarget<wxListCtrlEx>::OnEnter(x, y, def);
+		def = CScrollableDropTarget::OnEnter(x, y, def);
 		return OnDragOver(x, y, def);
 	}
 
@@ -316,10 +317,10 @@ BEGIN_EVENT_TABLE(CRemoteListView, CFileListCtrl<CGenericFileData>)
 	EVT_MENU(XRCID("ID_CONTEXT_REFRESH"), CRemoteListView::OnMenuRefresh)
 END_EVENT_TABLE()
 
-CRemoteListView::CRemoteListView(CView* pParent, CState& state, CQueueView* pQueue)
-	: CFileListCtrl<CGenericFileData>(pParent, pQueue),
-	CStateEventHandler(state),
-	m_parentView(pParent)
+CRemoteListView::CRemoteListView(CView* pParent, CState& state, CQueueView* pQueue, COptionsBase & options)
+	: CFileListCtrl<CGenericFileData>(pParent, pQueue, options)
+	, CStateEventHandler(state)
+	, m_parentView(pParent)
 {
 	state.RegisterHandler(this, STATECHANGE_REMOTE_DIR);
 	state.RegisterHandler(this, STATECHANGE_APPLYFILTER);
@@ -360,8 +361,8 @@ CRemoteListView::CRemoteListView(CView* pParent, CState& state, CQueueView* pQue
 
 CRemoteListView::~CRemoteListView()
 {
-	wxString str = wxString::Format(_T("%d %d"), m_sortDirection, m_sortColumn);
-	COptions::Get()->set(OPTION_REMOTEFILELIST_SORTORDER, str.ToStdWstring());
+	auto str = fz::sprintf(L"%d %d", m_sortDirection, m_sortColumn);
+	options_.set(OPTION_REMOTEFILELIST_SORTORDER, str);
 }
 
 // See comment to OnGetItemText
@@ -369,16 +370,23 @@ int CRemoteListView::OnGetItemImage(long item) const
 {
 	CRemoteListView *pThis = const_cast<CRemoteListView *>(this);
 	int index = GetItemIndex(item);
-	if (index == -1) {
+	if (index < 0) {
 		return -1;
 	}
 
+	if (index >= pThis->m_fileData.size()) {
+		return -1;
+	}
 	int &icon = pThis->m_fileData[index].icon;
 
 	if (icon != -2) {
 		return icon;
 	}
 
+	if (!m_pDirectoryListing || index >= m_pDirectoryListing->size()) {
+		return -1;
+	}
+
 	icon = pThis->GetIconIndex(iconType::file, (*m_pDirectoryListing)[index].name, false, (*m_pDirectoryListing)[index].is_dir());
 	return icon;
 }
@@ -415,6 +423,7 @@ void CRemoteListView::UpdateDirectoryListing_Added(std::shared_ptr<CDirectoryLis
 {
 	size_t const to_add = pDirectoryListing->size() - m_pDirectoryListing->size();
 	m_pDirectoryListing = pDirectoryListing;
+	UpdateSortComparisonObject();
 
 	m_indexMapping[0] = pDirectoryListing->size();
 
@@ -431,7 +440,7 @@ void CRemoteListView::UpdateDirectoryListing_Added(std::shared_ptr<CDirectoryLis
 		added_indexes.reserve(to_add);
 	}
 
-	std::unique_ptr<CFileListCtrlSortBase> compare = GetSortComparisonObject();
+	auto& compare = GetSortComparisonObject();
 	for (size_t i = pDirectoryListing->size() - to_add; i < pDirectoryListing->size(); ++i) {
 		CDirentry const& entry = (*pDirectoryListing)[i];
 		CGenericFileData data;
@@ -495,6 +504,7 @@ void CRemoteListView::UpdateDirectoryListing_Removed(std::shared_ptr<CDirectoryL
 	size_t const countRemoved = m_pDirectoryListing->size() - pDirectoryListing->size();
 	if (!countRemoved) {
 		m_pDirectoryListing = pDirectoryListing;
+		UpdateSortComparisonObject();
 		return;
 	}
 	wxASSERT(!IsComparing());
@@ -618,6 +628,7 @@ void CRemoteListView::UpdateDirectoryListing_Removed(std::shared_ptr<CDirectoryL
 	wxASSERT(m_indexMapping.size() == pDirectoryListing->size() + 1);
 
 	m_pDirectoryListing = pDirectoryListing;
+	UpdateSortComparisonObject();
 
 	if (m_pFilelistStatusBar) {
 		m_pFilelistStatusBar->SetHidden(m_pDirectoryListing->size() + 1 - m_indexMapping.size());
@@ -657,6 +668,7 @@ bool CRemoteListView::UpdateDirectoryListing(std::shared_ptr<CDirectoryListing>
 		}
 
 		m_pDirectoryListing = pDirectoryListing;
+		UpdateSortComparisonObject();
 
 		// We don't have to do anything
 		return true;
@@ -733,6 +745,7 @@ void CRemoteListView::SetDirectoryListing(std::shared_ptr<CDirectoryListing> con
 	}
 
 	m_pDirectoryListing = pDirectoryListing;
+	UpdateSortComparisonObject();
 
 	m_fileData.clear();
 	m_indexMapping.clear();
@@ -747,12 +760,16 @@ void CRemoteListView::SetDirectoryListing(std::shared_ptr<CDirectoryListing> con
 	if (m_pDirectoryListing) {
 		SetInfoText();
 
+		CFilterManager const& filter = m_state.GetStateFilterManager();
+		if (!filter.HasActiveRemoteFilters()) {
+			m_indexMapping.reserve(m_pDirectoryListing->size() + 1);
+			m_fileData.reserve(m_pDirectoryListing->size() + 1);
+		}
+
 		m_indexMapping.push_back(m_pDirectoryListing->size());
 
 		std::wstring const path = m_pDirectoryListing->path.GetPath();
 
-		CFilterManager const& filter = m_state.GetStateFilterManager();
-		
 		for (unsigned int i = 0; i < m_pDirectoryListing->size(); ++i) {
 			const CDirentry& entry = (*m_pDirectoryListing)[i];
 			CGenericFileData data;
@@ -764,7 +781,7 @@ void CRemoteListView::SetDirectoryListing(std::shared_ptr<CDirectoryListing> con
 				}
 #endif
 			}
-			m_fileData.push_back(data);
+			m_fileData.emplace_back(std::move(data));
 
 			if (filter.FilenameFiltered(entry.name, path, entry.is_dir(), entry.size, false, 0, entry.time)) {
 				++hidden;
@@ -789,7 +806,7 @@ void CRemoteListView::SetDirectoryListing(std::shared_ptr<CDirectoryListing> con
 
 		CGenericFileData data;
 		data.icon = m_dirIcon;
-		m_fileData.push_back(data);
+		m_fileData.emplace_back(std::move(data));
 	}
 	else {
 		eraseBackground = true;
@@ -847,36 +864,9 @@ void CRemoteListView::SetDirectoryListing(std::shared_ptr<CDirectoryListing> con
 	}
 }
 
-// Filenames on VMS systems have a revision suffix, e.g.
-// foo.bar;1
-// foo.bar;2
-// foo.bar;3
-std::wstring StripVMSRevision(std::wstring const& name)
-{
-	size_t pos = name.rfind(';');
-	if (pos == std::wstring::npos || !pos) {
-		return name;
-	}
-
-	if (pos == name.size() - 1) {
-		return name;
-	}
-
-	size_t p = pos;
-	while (++p < name.size()) {
-		wchar_t const& c = name[p];
-		if (c < '0' || c > '9') {
-			return name;
-		}
-	}
-
-	return name.substr(0, pos);
-}
-
-
 void CRemoteListView::OnItemActivated(wxListEvent &event)
 {
-	int const action = COptions::Get()->get_int(OPTION_DOUBLECLICK_ACTION_DIRECTORY);
+	int const action = options_.get_int(OPTION_DOUBLECLICK_ACTION_DIRECTORY);
 	if (!m_state.IsRemoteIdle(action ? false : true)) {
 		wxBell();
 		return;
@@ -961,7 +951,7 @@ void CRemoteListView::OnItemActivated(wxListEvent &event)
 			}
 		}
 		else {
-			const int action = COptions::Get()->get_int(OPTION_DOUBLECLICK_ACTION_FILE);
+			const int action = options_.get_int(OPTION_DOUBLECLICK_ACTION_FILE);
 			if (action == 3) {
 				// No action
 				wxBell();
@@ -984,7 +974,7 @@ void CRemoteListView::OnItemActivated(wxListEvent &event)
 			}
 
 			std::wstring localFile = CQueueView::ReplaceInvalidCharacters(name);
-			if (m_pDirectoryListing->path.GetType() == VMS && COptions::Get()->get_int(OPTION_STRIP_VMS_REVISION)) {
+			if (m_pDirectoryListing->path.GetType() == VMS && options_.get_int(OPTION_STRIP_VMS_REVISION)) {
 				localFile = StripVMSRevision(localFile);
 			}
 			m_pQueue->QueueFile(queue_only, true, name,
@@ -1234,7 +1224,7 @@ void CRemoteListView::OnMenuDownload(wxCommandEvent& event)
 	TransferSelectedFiles(localDir, event.GetId() == XRCID("ID_ADDTOQUEUE"));
 }
 
-void CRemoteListView::TransferSelectedFiles(const CLocalPath& local_parent, bool queue_only)
+void CRemoteListView::TransferSelectedFiles(const CLocalPath& local_parent, bool queue_only, transfer_flags custom_flags, transfer_flags custom_flags_mask)
 {
 	bool idle = m_state.IsRemoteIdle();
 
@@ -1286,12 +1276,13 @@ void CRemoteListView::TransferSelectedFiles(const CLocalPath& local_parent, bool
 		}
 		else {
 			std::wstring localFile = CQueueView::ReplaceInvalidCharacters(name);
-			if (m_pDirectoryListing->path.GetType() == VMS && COptions::Get()->get_int(OPTION_STRIP_VMS_REVISION)) {
+			if (m_pDirectoryListing->path.GetType() == VMS && options_.get_int(OPTION_STRIP_VMS_REVISION)) {
 				localFile = StripVMSRevision(localFile);
 			}
 			m_pQueue->QueueFile(queue_only, true,
 				name, (name == localFile) ? std::wstring() : localFile,
-				local_parent, m_pDirectoryListing->path, site, entry.size);
+				local_parent, m_pDirectoryListing->path, site, entry.size, CEditHandler::none,
+				QueuePriority::normal, custom_flags, custom_flags_mask);
 			added = true;
 		}
 	}
@@ -1302,7 +1293,7 @@ void CRemoteListView::TransferSelectedFiles(const CLocalPath& local_parent, bool
 	if (!root.empty()) {
 		pRecursiveOperation->AddRecursionRoot(std::move(root));
 		CFilterManager filter;
-		pRecursiveOperation->StartRecursiveOperation(CRecursiveOperation::recursive_transfer, filter.GetActiveFilters(), !queue_only);
+		pRecursiveOperation->StartRecursiveOperation(recursive_operation::recursive_transfer, filter.GetActiveFilters(), !queue_only);
 	}
 }
 
@@ -1368,7 +1359,8 @@ CServerPath CRemoteListView::MenuMkdir()
 		return CServerPath();
 	}
 
-	m_state.m_pCommandQueue->ProcessCommand(new CMkdirCommand(path));
+	transfer_flags const flags = GetMkdirFlags(m_state.GetSite().server, options_, path);
+	m_state.m_pCommandQueue->ProcessCommand(new CMkdirCommand(path, flags));
 
 	// Return name of the New Directory
 	return path;
@@ -1533,7 +1525,7 @@ void CRemoteListView::OnMenuDelete(wxCommandEvent&)
 
 		if (!root.empty()) {
 			pRecursiveOperation->AddRecursionRoot(std::move(root));
-			pRecursiveOperation->StartRecursiveOperation(CRecursiveOperation::recursive_delete, filter.GetActiveFilters());
+			pRecursiveOperation->StartRecursiveOperation(recursive_operation::recursive_delete, filter.GetActiveFilters());
 		}
 	}
 }
@@ -1697,7 +1689,7 @@ void CRemoteListView::OnMenuChmod(wxCommandEvent&)
 {
 	Site const& site = m_state.GetSite();
 	auto protocol = site.server.GetProtocol();
-	
+
 	if (!m_state.IsRemoteConnected() || !m_state.IsRemoteIdle()) {
 		wxBell();
 		return;
@@ -1844,11 +1836,11 @@ void CRemoteListView::HandleGenericChmod(ChmodUICommand &command)
 		pRecursiveOperation->SetChmodData(std::move(chmodData));
 		pRecursiveOperation->AddRecursionRoot(std::move(root));
 		CFilterManager filter;
-		pRecursiveOperation->StartRecursiveOperation(CRecursiveOperation::recursive_chmod, filter.GetActiveFilters());
+		pRecursiveOperation->StartRecursiveOperation(recursive_operation::recursive_chmod, filter.GetActiveFilters());
 
-		// Refresh listing. This gets done implicitely by the recursive operation, so
+		// Refresh listing. This gets done implicitly by the recursive operation, so
 		// only it if not recursing.
-		if (pRecursiveOperation->GetOperationMode() != CRecursiveOperation::recursive_chmod) {
+		if (pRecursiveOperation->GetOperationMode() != recursive_operation::recursive_chmod) {
 			m_state.ChangeRemoteDir(m_pDirectoryListing->path);
 		}
 	}
@@ -1895,7 +1887,7 @@ void CRemoteListView::ApplyCurrentFilter()
 			++hidden;
 			continue;
 		}
-	
+
 		if (entry.is_dir()) {
 			++totalDirCount;
 		}
@@ -1947,7 +1939,7 @@ std::vector<std::wstring> CRemoteListView::RememberSelectedItems(std::wstring& f
 			if (item < 0) {
 				break;
 			}
-			
+
 			if (!item) {
 				selectedNames.push_back(L"..");
 			}
@@ -2119,10 +2111,10 @@ void CRemoteListView::OnStateChange(t_statechange_notifications notification, st
 	}
 	else if (notification == STATECHANGE_SERVER) {
 		if (m_windowTinter) {
-			m_windowTinter->SetBackgroundTint(m_state.GetSite().m_colour);
+			m_windowTinter->SetBackgroundTint(site_colour_to_wx(m_state.GetSite().m_colour));
 		}
 		if (m_pInfoText) {
-			m_pInfoText->SetBackgroundTint(m_state.GetSite().m_colour);
+			m_pInfoText->SetBackgroundTint(site_colour_to_wx(m_state.GetSite().m_colour));
 		}
 	}
 	else {
@@ -2163,7 +2155,7 @@ void CRemoteListView::SetInfoText()
 
 void CRemoteListView::OnBeginDrag(wxListEvent&)
 {
-	if (COptions::Get()->get_int(OPTION_DND_DISABLED) != 0) {
+	if (options_.get_int(OPTION_DND_DISABLED) != 0) {
 		return;
 	}
 
@@ -2224,6 +2216,7 @@ void CRemoteListView::OnBeginDrag(wxListEvent&)
 	pDragDropManager->pDragSource = this;
 	pDragDropManager->site = site;
 	pDragDropManager->remoteParent = m_pDirectoryListing->path;
+	pDragDropManager->dragDataObject = pRemoteDataObject;
 
 	// Add files to remote data object
 	item = -1;
@@ -2262,10 +2255,9 @@ void CRemoteListView::OnBeginDrag(wxListEvent&)
 
 	CLabelEditBlocker b(*this);
 
-	wxDropSource source(this);
+	DropSource source(this);
 	source.SetData(object);
-
-	int res = source.DoDragDrop();
+	int res = source.DoFileDragDrop();
 
 	pDragDropManager->Release();
 
@@ -2273,9 +2265,13 @@ void CRemoteListView::OnBeginDrag(wxListEvent&)
 		return;
 	}
 
+#if FZ3_USESHELLEXT || defined(__WXMAC__)
 #if FZ3_USESHELLEXT
 	if (ext) {
 		if (!pRemoteDataObject->DidSendData()) {
+#else
+	if (!source.m_OutDir.IsEmpty()) {
+#endif
 			Site newSite = m_state.GetSite();
 			if (!m_state.IsRemoteIdle() ||
 				!newSite || newSite.server != pRemoteDataObject->GetSite().server ||
@@ -2312,16 +2308,26 @@ void CRemoteListView::OnBeginDrag(wxListEvent&)
 					return;
 				}
 			}
-
+#if FZ3_USESHELLEXT
 			CLocalPath target(ext->GetTarget().ToStdWstring());
+#else
+			// wxMessageBoxEx(source.m_OutDir);
+			CLocalPath target(source.m_OutDir.ToStdWstring());
+#endif
 			if (target.empty()) {
+#if FZ3_USESHELLEXT
 				ext.reset(); // Release extension before the modal message box
 				wxMessageBoxEx(_("Could not determine the target of the Drag&Drop operation.\nEither the shell extension is not installed properly or you didn't drop the files into an Explorer window."));
+#else
+				wxMessageBoxEx(_("Could not determine the target of the Drag&Drop operation."));
+#endif
 				return;
 			}
 
 			TransferSelectedFiles(target, false);
+#if FZ3_USESHELLEXT
 		}
+#endif
 	}
 #endif
 }
@@ -2506,10 +2512,10 @@ wxString CRemoteListView::GetItemText(int item, unsigned int column)
 	}
 
 	if (!column) {
-		if ((size_t)index == m_pDirectoryListing->size()) {
+		if (m_pDirectoryListing && (size_t)index == m_pDirectoryListing->size()) {
 			return _T("..");
 		}
-		else if ((size_t)index < m_pDirectoryListing->size()) {
+		else if (m_pDirectoryListing && (size_t)index < m_pDirectoryListing->size()) {
 			return (*m_pDirectoryListing)[index].name;
 		}
 		else {
@@ -2520,7 +2526,7 @@ wxString CRemoteListView::GetItemText(int item, unsigned int column)
 		return wxString(); //.. has no attributes
 	}
 
-	if ((size_t)index >= m_pDirectoryListing->size()) {
+	if (!m_pDirectoryListing || (size_t)index >= m_pDirectoryListing->size()) {
 		return wxString();
 	}
 
@@ -2560,50 +2566,51 @@ wxString CRemoteListView::GetItemText(int item, unsigned int column)
 	return wxString();
 }
 
-std::unique_ptr<CFileListCtrlSortBase> CRemoteListView::GetSortComparisonObject()
+void CRemoteListView::UpdateSortComparisonObject()
 {
 	CFileListCtrlSort<CDirectoryListing>::DirSortMode dirSortMode = GetDirSortMode();
 	NameSortMode nameSortMode = GetNameSortMode();
 
-	CDirectoryListing const& directoryListing = *m_pDirectoryListing;
+	static CDirectoryListing const empty;
+	CDirectoryListing const& directoryListing = m_pDirectoryListing ? *m_pDirectoryListing : empty;
 	if (!m_sortDirection) {
 		if (m_sortColumn == 1) {
-			return std::make_unique<CFileListCtrlSortSize<CDirectoryListing, CGenericFileData>>(directoryListing, m_fileData, dirSortMode, nameSortMode, this);
+			sortComparisonObject_ = std::make_unique<CFileListCtrlSortSize<CDirectoryListing, CGenericFileData>>(directoryListing, m_fileData, dirSortMode, nameSortMode, this);
 		}
 		else if (m_sortColumn == 2) {
-			return std::make_unique<CFileListCtrlSortType<CDirectoryListing, CGenericFileData>>(directoryListing, m_fileData, dirSortMode, nameSortMode, this);
+			sortComparisonObject_ = std::make_unique<CFileListCtrlSortType<CDirectoryListing, CGenericFileData>>(directoryListing, m_fileData, dirSortMode, nameSortMode, this);
 		}
 		else if (m_sortColumn == 3) {
-			return std::make_unique<CFileListCtrlSortTime<CDirectoryListing, CGenericFileData>>(directoryListing, m_fileData, dirSortMode, nameSortMode, this);
+			sortComparisonObject_ = std::make_unique<CFileListCtrlSortTime<CDirectoryListing, CGenericFileData>>(directoryListing, m_fileData, dirSortMode, nameSortMode, this);
 		}
 		else if (m_sortColumn == 4) {
-			return std::make_unique<CFileListCtrlSortPermissions<CDirectoryListing, CGenericFileData>>(directoryListing, m_fileData, dirSortMode, nameSortMode, this);
+			sortComparisonObject_ = std::make_unique<CFileListCtrlSortPermissions<CDirectoryListing, CGenericFileData>>(directoryListing, m_fileData, dirSortMode, nameSortMode, this);
 		}
 		else if (m_sortColumn == 5) {
-			return std::make_unique<CFileListCtrlSortOwnerGroup<CDirectoryListing, CGenericFileData>>(directoryListing, m_fileData, dirSortMode, nameSortMode, this);
+			sortComparisonObject_ = std::make_unique<CFileListCtrlSortOwnerGroup<CDirectoryListing, CGenericFileData>>(directoryListing, m_fileData, dirSortMode, nameSortMode, this);
 		}
 		else {
-			return std::make_unique<CFileListCtrlSortName<CDirectoryListing, CGenericFileData>>(directoryListing, m_fileData, dirSortMode, nameSortMode, this);
+			sortComparisonObject_ = std::make_unique<CFileListCtrlSortName<CDirectoryListing, CGenericFileData>>(directoryListing, m_fileData, dirSortMode, nameSortMode, this);
 		}
 	}
 	else {
 		if (m_sortColumn == 1) {
-			return std::make_unique<CReverseSort<CFileListCtrlSortSize<CDirectoryListing, CGenericFileData>, CGenericFileData>>(directoryListing, m_fileData, dirSortMode, nameSortMode, this);
+			sortComparisonObject_ = std::make_unique<CReverseSort<CFileListCtrlSortSize<CDirectoryListing, CGenericFileData>, CGenericFileData>>(directoryListing, m_fileData, dirSortMode, nameSortMode, this);
 		}
 		else if (m_sortColumn == 2) {
-			return std::make_unique<CReverseSort<CFileListCtrlSortType<CDirectoryListing, CGenericFileData>, CGenericFileData>>(directoryListing, m_fileData, dirSortMode, nameSortMode, this);
+			sortComparisonObject_ = std::make_unique<CReverseSort<CFileListCtrlSortType<CDirectoryListing, CGenericFileData>, CGenericFileData>>(directoryListing, m_fileData, dirSortMode, nameSortMode, this);
 		}
 		else if (m_sortColumn == 3) {
-			return std::make_unique<CReverseSort<CFileListCtrlSortTime<CDirectoryListing, CGenericFileData>, CGenericFileData>>(directoryListing, m_fileData, dirSortMode, nameSortMode, this);
+			sortComparisonObject_ = std::make_unique<CReverseSort<CFileListCtrlSortTime<CDirectoryListing, CGenericFileData>, CGenericFileData>>(directoryListing, m_fileData, dirSortMode, nameSortMode, this);
 		}
 		else if (m_sortColumn == 4) {
-			return std::make_unique<CReverseSort<CFileListCtrlSortPermissions<CDirectoryListing, CGenericFileData>, CGenericFileData>>(directoryListing, m_fileData, dirSortMode, nameSortMode, this);
+			sortComparisonObject_ = std::make_unique<CReverseSort<CFileListCtrlSortPermissions<CDirectoryListing, CGenericFileData>, CGenericFileData>>(directoryListing, m_fileData, dirSortMode, nameSortMode, this);
 		}
 		else if (m_sortColumn == 5) {
-			return std::make_unique<CReverseSort<CFileListCtrlSortOwnerGroup<CDirectoryListing, CGenericFileData>, CGenericFileData>>(directoryListing, m_fileData, dirSortMode, nameSortMode, this);
+			sortComparisonObject_ = std::make_unique<CReverseSort<CFileListCtrlSortOwnerGroup<CDirectoryListing, CGenericFileData>, CGenericFileData>>(directoryListing, m_fileData, dirSortMode, nameSortMode, this);
 		}
 		else {
-			return std::make_unique<CReverseSort<CFileListCtrlSortName<CDirectoryListing, CGenericFileData>, CGenericFileData>>(directoryListing, m_fileData, dirSortMode, nameSortMode, this);
+			sortComparisonObject_ = std::make_unique<CReverseSort<CFileListCtrlSortName<CDirectoryListing, CGenericFileData>, CGenericFileData>>(directoryListing, m_fileData, dirSortMode, nameSortMode, this);
 		}
 	}
 }
@@ -2628,7 +2635,7 @@ void CRemoteListView::LinkIsNotDir(CServerPath const& path, std::wstring const&
 {
 	if (m_pLinkResolveState && m_pLinkResolveState->remote_path == path && m_pLinkResolveState->link == link) {
 		std::wstring localFile = CQueueView::ReplaceInvalidCharacters(link);
-		if (m_pDirectoryListing->path.GetType() == VMS && COptions::Get()->get_int(OPTION_STRIP_VMS_REVISION)) {
+		if (m_pDirectoryListing->path.GetType() == VMS && options_.get_int(OPTION_STRIP_VMS_REVISION)) {
 			localFile = StripVMSRevision(localFile);
 		}
 		m_pQueue->QueueFile(false, true,
@@ -2806,6 +2813,6 @@ void CRemoteListView::OnMenuNewfile(wxCommandEvent&)
 		return;
 	}
 
-	CFileTransferCommand *cmd = new CFileTransferCommand(emptyfile, m_pDirectoryListing->path, newFileName, transfer_flags());
+	CFileTransferCommand *cmd = new CFileTransferCommand(memory_reader_factory(L"Empty file", std::string_view()), m_pDirectoryListing->path, newFileName, transfer_flags());
 	m_state.m_pCommandQueue->ProcessCommand(cmd);
 }
