@@ -16,12 +16,12 @@ class CRemoteListView final : public CFileListCtrl<CGenericFileData>, CStateEven
 {
 	friend class CRemoteListViewDropTarget;
 public:
-	CRemoteListView(CView* pParent, CState& state, CQueueView* pQueue);
+	CRemoteListView(CView* pParent, CState& state, CQueueView* pQueue, COptionsBase & options);
 	virtual ~CRemoteListView();
 
 	virtual bool CanStartComparison();
 	virtual void StartComparison();
-	virtual bool get_next_file(std::wstring & name, bool &dir, int64_t &size, fz::datetime& date) override;
+	virtual bool get_next_file(std::wstring_view & name, std::wstring & path, bool &dir, int64_t &size, fz::datetime& date) override;
 	virtual void FinishComparison();
 	virtual void OnExitComparisonMode();
 
@@ -65,7 +65,7 @@ protected:
 	bool IsItemValid(unsigned int item) const;
 	int GetItemIndex(unsigned int item) const;
 
-	virtual std::unique_ptr<CFileListCtrlSortBase> GetSortComparisonObject() override;
+	virtual void UpdateSortComparisonObject() override;
 
 	virtual void OnStateChange(t_statechange_notifications notification, std::wstring const& data, const void* data2) override;
 	void ApplyCurrentFilter();
@@ -83,15 +83,14 @@ protected:
 	std::shared_ptr<CDirectoryListing> m_pDirectoryListing;
 
 	// Caller is responsible to check selection is valid!
-	void TransferSelectedFiles(const CLocalPath& local_parent, bool queue_only);
+	void TransferSelectedFiles(const CLocalPath& local_parent, bool queue_only, transfer_flags custom_flags = transfer_flags::none, transfer_flags custom_flags_mask = transfer_flags::none);
 
 	void HandleGenericChmod(ChmodUICommand &command);
 
 	// Cache icon for directories, no need to calculate it multiple times
 	int m_dirIcon;
 
-	CInfoText* m_pInfoText;
-	void RepositionInfoText();
+	CInfoText* m_pInfoText{};
 	void SetInfoText();
 
 	virtual bool OnBeginRename(const wxListEvent& event);
@@ -120,7 +119,7 @@ protected:
 	std::unique_ptr<CWindowTinter> m_windowTinter;
 
 	CView *m_parentView{};
-	
+
 	DECLARE_EVENT_TABLE()
 	void OnItemActivated(wxListEvent &event);
 	void OnContextMenu(wxContextMenuEvent& event);
