@@ -1,4 +1,4 @@
-#include <filezilla.h>
+#include "filezilla.h"
 
 #include "chmoddialog.h"
 #include "commandqueue.h"
@@ -13,6 +13,7 @@
 #include "RemoteTreeView.h"
 #include "remote_recursive_operation.h"
 #include "themeprovider.h"
+#include "../commonui/misc.h"
 
 #include <libfilezilla/uri.hpp>
 
@@ -21,7 +22,6 @@
 
 #include <algorithm>
 
-
 class CItemData : public wxTreeItemData
 {
 public:
@@ -38,15 +38,6 @@ public:
 	{
 	}
 
-	void ClearDropHighlight()
-	{
-		const wxTreeItemId dropHighlight = m_pRemoteTreeView->m_dropHighlight;
-		if (dropHighlight != wxTreeItemId()) {
-			m_pRemoteTreeView->SetItemDropHighlight(dropHighlight, false);
-			m_pRemoteTreeView->m_dropHighlight = wxTreeItemId();
-		}
-	}
-
 	wxTreeItemId GetHit(const wxPoint& point)
 	{
 		int flags = 0;
@@ -87,34 +78,35 @@ public:
 			pDragDropManager->pDropTarget = m_pRemoteTreeView;
 		}
 
-		auto const format = m_pDataObject->GetReceivedFormat();
+		auto const format = GetReceivedFormat();
 		if (format == m_pFileDataObject->GetFormat()) {
 			m_pRemoteTreeView->m_state.UploadDroppedFiles(m_pFileDataObject, path, false);
 		}
-		else if (format == m_pLocalDataObject->GetFormat()) {
-			m_pRemoteTreeView->m_state.UploadDroppedFiles(m_pLocalDataObject, path, false);
+		else if (format == LocalDataObjectFormat()) {
+			m_pRemoteTreeView->m_state.UploadDroppedFiles(GetLocalDataObject(), path, false);
 		}
 		else {
-			if (m_pRemoteDataObject->GetProcessId() != (int)wxGetProcessId()) {
+			auto * obj = GetRemoteDataObject();
+			if (obj->GetProcessId() != (int)wxGetProcessId()) {
 				wxMessageBoxEx(_("Drag&drop between different instances of FileZilla has not been implemented yet."));
 				return wxDragNone;
 			}
 
-			if (!m_pRemoteTreeView->m_state.GetSite() || m_pRemoteDataObject->GetSite().server != m_pRemoteTreeView->m_state.GetSite().server) {
+			if (!m_pRemoteTreeView->m_state.GetSite() || obj->GetSite().server != m_pRemoteTreeView->m_state.GetSite().server) {
 				wxMessageBoxEx(_("Drag&drop between different servers has not been implemented yet."));
 				return wxDragNone;
 			}
 
 			// Make sure path path is valid
-			if (path == m_pRemoteDataObject->GetServerPath()) {
+			if (path == obj->GetServerPath()) {
 				wxMessageBoxEx(_("Source and path of the drop operation are identical"));
 				return wxDragNone;
 			}
 
-			std::vector<CRemoteDataObject::t_fileInfo> const& files = m_pRemoteDataObject->GetFiles();
+			std::vector<CRemoteDataObject::t_fileInfo> const& files = obj->GetFiles();
 			for (auto const& info : files) {
 				if (info.dir) {
-					CServerPath dir = m_pRemoteDataObject->GetServerPath();
+					CServerPath dir = obj->GetServerPath();
 					dir.AddSegment(info.name);
 					if (dir == path) {
 						return wxDragNone;
@@ -128,7 +120,7 @@ public:
 
 			for (auto const& info : files) {
 				m_pRemoteTreeView->m_state.m_pCommandQueue->ProcessCommand(
-					new CRenameCommand(m_pRemoteDataObject->GetServerPath(), info.name, path, info.name)
+					new CRenameCommand(obj->GetServerPath(), info.name, path, info.name)
 					);
 			}
 
@@ -140,10 +132,10 @@ public:
 
 	virtual bool OnDrop(wxCoord x, wxCoord y)
 	{
-		if (!CScrollableDropTarget<wxTreeCtrlEx>::OnDrop(x, y)) {
+		if (!CScrollableDropTarget::OnDrop(x, y)) {
 			return false;
 		}
-		ClearDropHighlight();
+		m_pRemoteTreeView->ClearDropHighlight();
 
 		wxTreeItemId hit = GetHit(wxPoint(x, y));
 		if (!hit) {
@@ -158,41 +150,35 @@ public:
 		return true;
 	}
 
-	wxTreeItemId DisplayDropHighlight(wxPoint point)
+	wxTreeItemId DisplayDropHighlight(wxPoint const& point) override
 	{
 		wxTreeItemId hit = GetHit(point);
 		if (!hit) {
-			ClearDropHighlight();
+			m_pRemoteTreeView->ClearDropHighlight();
 			return wxTreeItemId();
 		}
 
-		const CServerPath& path = m_pRemoteTreeView->GetPathFromItem(hit);
+		CServerPath const& path = m_pRemoteTreeView->GetPathFromItem(hit);
 
 		if (path.empty()) {
-			ClearDropHighlight();
+			m_pRemoteTreeView->ClearDropHighlight();
 			return wxTreeItemId();
 		}
 
-		const wxTreeItemId dropHighlight = m_pRemoteTreeView->m_dropHighlight;
-		if (dropHighlight != wxTreeItemId()) {
-			m_pRemoteTreeView->SetItemDropHighlight(dropHighlight, false);
-		}
-
-		m_pRemoteTreeView->SetItemDropHighlight(hit, true);
-		m_pRemoteTreeView->m_dropHighlight = hit;
+		m_pRemoteTreeView->DisplayDropHighlight(hit);
 
 		return hit;
 	}
 
 	virtual wxDragResult OnDragOver(wxCoord x, wxCoord y, wxDragResult def)
 	{
-		def = CScrollableDropTarget<wxTreeCtrlEx>::OnDragOver(x, y, def);
+		def = CScrollableDropTarget::OnDragOver(x, y, def);
 
 		if (def == wxDragError ||
 			def == wxDragNone ||
 			def == wxDragCancel)
 		{
-			ClearDropHighlight();
+			m_pRemoteTreeView->ClearDropHighlight();
 			return def;
 		}
 
@@ -210,13 +196,13 @@ public:
 
 	virtual void OnLeave()
 	{
-		CScrollableDropTarget<wxTreeCtrlEx>::OnLeave();
-		ClearDropHighlight();
+		CScrollableDropTarget::OnLeave();
+		m_pRemoteTreeView->ClearDropHighlight();
 	}
 
 	virtual wxDragResult OnEnter(wxCoord x, wxCoord y, wxDragResult def)
 	{
-		def = CScrollableDropTarget<wxTreeCtrlEx>::OnEnter(x, y, def);
+		def = CScrollableDropTarget::OnEnter(x, y, def);
 		return OnDragOver(x, y, def);
 	}
 
@@ -244,9 +230,10 @@ EVT_MENU(XRCID("ID_GETURL"), CRemoteTreeView::OnMenuGeturl)
 END_EVENT_TABLE()
 
 CRemoteTreeView::CRemoteTreeView(wxWindow* parent, wxWindowID id, CState& state, CQueueView* pQueue)
-	: wxTreeCtrlEx(parent, id, wxDefaultPosition, wxDefaultSize, DEFAULT_TREE_STYLE | wxTAB_TRAVERSAL | wxTR_EDIT_LABELS | wxNO_BORDER | wxTR_HIDE_ROOT),
-	CSystemImageList(CThemeProvider::GetIconSize(iconSizeSmall).x),
-	CStateEventHandler(state)
+	: wxTreeCtrlEx(parent, id, wxDefaultPosition, wxDefaultSize, DEFAULT_TREE_STYLE | wxTAB_TRAVERSAL | wxTR_EDIT_LABELS | wxNO_BORDER | wxTR_HIDE_ROOT)
+	, CSystemImageList(CThemeProvider::GetIconSize(iconSizeSmall).x)
+	, CStateEventHandler(state)
+	, COptionChangeEventHandler(this)
 {
 #ifdef __WXMAC__
 	SetFont(wxSystemSettings::GetFont(wxSYS_DEFAULT_GUI_FONT));
@@ -259,7 +246,7 @@ CRemoteTreeView::CRemoteTreeView(wxWindow* parent, wxWindowID id, CState& state,
 	CreateImageList();
 
 	UpdateSortMode();
-	RegisterOption(OPTION_FILELIST_NAMESORT);
+	COptions::Get()->watch(OPTION_FILELIST_NAMESORT, this);
 
 	m_pQueue = pQueue;
 	AddRoot(_T(""));
@@ -274,6 +261,7 @@ CRemoteTreeView::CRemoteTreeView(wxWindow* parent, wxWindowID id, CState& state,
 
 CRemoteTreeView::~CRemoteTreeView()
 {
+	COptions::Get()->unwatch_all(this);
 	SetImageList(0);
 	delete m_pImageList;
 }
@@ -287,7 +275,7 @@ void CRemoteTreeView::OnStateChange(t_statechange_notifications notification, st
 		ApplyFilters(false);
 	}
 	else if (notification == STATECHANGE_SERVER) {
-		m_windowTinter->SetBackgroundTint(m_state.GetSite().m_colour);
+		m_windowTinter->SetBackgroundTint(site_colour_to_wx(m_state.GetSite().m_colour));
 	}
 }
 
@@ -418,7 +406,7 @@ wxTreeItemId CRemoteTreeView::MakeParent(CServerPath path, bool select)
 		if (!child) {
 			CDirectoryListing listing;
 
-			if (m_state.m_pEngine->CacheLookup(path, listing) == FZ_REPLY_OK) {
+			if (m_state.engine_->CacheLookup(path, listing) == FZ_REPLY_OK) {
 				child = AppendItem(parent, *iter, 0, 2, path.HasParent() ? 0 : new CItemData(path));
 				SetItemImages(child, false);
 			}
@@ -535,7 +523,7 @@ void CRemoteTreeView::DisplayItem(wxTreeItemId parent, const CDirectoryListing&
 
 	std::wstring const path = listing.path.GetPath();
 
-	CFilterDialog filter;
+	CFilterManager const& filter = m_state.GetStateFilterManager();
 
 	wxTreeItemId last;
 	for (size_t i = 0; i < listing.size(); ++i) {
@@ -554,7 +542,7 @@ void CRemoteTreeView::DisplayItem(wxTreeItemId parent, const CDirectoryListing&
 
 		CDirectoryListing subListing;
 
-		if (m_state.m_pEngine->CacheLookup(subdir, subListing) == FZ_REPLY_OK) {
+		if (m_state.engine_->CacheLookup(subdir, subListing) == FZ_REPLY_OK) {
 			wxTreeItemId child = AppendItem(parent, name, 0, 2, 0);
 			SetItemImages(child, false);
 
@@ -602,7 +590,7 @@ void CRemoteTreeView::RefreshItem(wxTreeItemId parent, const CDirectoryListing&
 	std::vector<wxTreeItemId> toDelete;
 
 	bool inserted = false;
-	
+
 	wxTreeItemIdValue unused;
 	child = GetFirstChild(parent, unused);
 
@@ -618,7 +606,7 @@ void CRemoteTreeView::RefreshItem(wxTreeItemId parent, const CDirectoryListing&
 			childPath.AddSegment(*iter);
 
 			CDirectoryListing subListing;
-			if (m_state.m_pEngine->CacheLookup(childPath, subListing) == FZ_REPLY_OK) {
+			if (m_state.engine_->CacheLookup(childPath, subListing) == FZ_REPLY_OK) {
 				if (!GetLastChild(child) && HasSubdirs(subListing, filter)) {
 					AppendItem(child, _T(""), -1, -1);
 				}
@@ -648,7 +636,7 @@ void CRemoteTreeView::RefreshItem(wxTreeItemId parent, const CDirectoryListing&
 			childPath.AddSegment(*iter);
 
 			CDirectoryListing subListing;
-			if (m_state.m_pEngine->CacheLookup(childPath, subListing) == FZ_REPLY_OK) {
+			if (m_state.engine_->CacheLookup(childPath, subListing) == FZ_REPLY_OK) {
 				last = InsertItem(parent, last, *iter, 0, 2, 0);
 				if (last) {
 					SetItemImages(last, false);
@@ -685,7 +673,7 @@ void CRemoteTreeView::RefreshItem(wxTreeItemId parent, const CDirectoryListing&
 		childPath.AddSegment(*iter);
 
 		CDirectoryListing subListing;
-		if (m_state.m_pEngine->CacheLookup(childPath, subListing) == FZ_REPLY_OK) {
+		if (m_state.engine_->CacheLookup(childPath, subListing) == FZ_REPLY_OK) {
 			last = InsertItem(parent, last, *iter, 0, 2, 0);
 			if (last) {
 				SetItemImages(last, false);
@@ -816,7 +804,7 @@ CServerPath CRemoteTreeView::GetPathFromItem(const wxTreeItemId& item) const
 
 void CRemoteTreeView::OnBeginDrag(wxTreeEvent& event)
 {
-	if (COptions::Get()->GetOptionVal(OPTION_DND_DISABLED) != 0) {
+	if (COptions::Get()->get_int(OPTION_DND_DISABLED) != 0) {
 		return;
 	}
 
@@ -874,11 +862,11 @@ void CRemoteTreeView::OnBeginDrag(wxTreeEvent& event)
 	pDragDropManager->pDragSource = this;
 	pDragDropManager->site = site;
 	pDragDropManager->remoteParent = parent;
+	pDragDropManager->dragDataObject = pRemoteDataObject;
 
-	wxDropSource source(this);
+	DropSource source(this);
 	source.SetData(object);
-
-	int res = source.DoDragDrop();
+	int res = source.DoFileDragDrop();
 
 	pDragDropManager->Release();
 
@@ -886,24 +874,39 @@ void CRemoteTreeView::OnBeginDrag(wxTreeEvent& event)
 		return;
 	}
 
+#if FZ3_USESHELLEXT || defined(__WXMAC__)
 #if FZ3_USESHELLEXT
 	if (ext) {
 		if (!pRemoteDataObject->DidSendData()) {
+#else
+	if (!source.m_OutDir.IsEmpty()) {
+#endif
 			Site newSite = m_state.GetSite();
 			if (!newSite || !m_state.IsRemoteIdle() || site != newSite) {
 				wxBell();
 				return;
 			}
 
+#if FZ3_USESHELLEXT
 			CLocalPath target(ext->GetTarget().ToStdWstring());
+#else
+			// wxMessageBoxEx(source.m_OutDir);
+			CLocalPath target(source.m_OutDir.ToStdWstring());
+#endif
 			if (target.empty()) {
+#if FZ3_USESHELLEXT
 				ext.reset(); // Release extension before the modal message box
 				wxMessageBoxEx(_("Could not determine the target of the Drag&Drop operation.\nEither the shell extension is not installed properly or you didn't drop the files into an Explorer window."));
+#else
+				wxMessageBoxEx(_("Could not determine the target of the Drag&Drop operation."));
+#endif
 				return;
 			}
 
 			m_state.DownloadDroppedFiles(pRemoteDataObject, target);
+#if FZ3_USESHELLEXT
 		}
+#endif
 	}
 #endif
 }
@@ -995,7 +998,7 @@ void CRemoteTreeView::OnMenuChmod(wxCommandEvent&)
 		CDirectoryListing listing;
 
 		// ... and it needs to be cached
-		cached = m_state.m_pEngine->CacheLookup(parentPath, listing) == FZ_REPLY_OK;
+		cached = m_state.engine_->CacheLookup(parentPath, listing) == FZ_REPLY_OK;
 		if (cached) {
 			for (size_t i = 0; i < listing.size(); ++i) {
 				auto const& entry = listing[i];
@@ -1053,20 +1056,8 @@ void CRemoteTreeView::OnMenuChmod(wxCommandEvent&)
 		chmodDialog.reset();
 		pRecursiveOperation->SetChmodData(std::move(chmodData));
 
-		CServerPath currentPath;
-		const wxTreeItemId selected = GetSelection();
-		if (selected) {
-			currentPath = GetPathFromItem(selected);
-		}
 		CFilterManager filter;
-		pRecursiveOperation->StartRecursiveOperation(CRecursiveOperation::recursive_chmod, filter.GetActiveFilters(), currentPath);
-	}
-	else {
-		const wxTreeItemId selected = GetSelection();
-		if (selected) {
-			CServerPath currentPath = GetPathFromItem(selected);
-			m_state.ChangeRemoteDir(currentPath);
-		}
+		pRecursiveOperation->StartRecursiveOperation(recursive_operation::recursive_chmod, filter.GetActiveFilters());
 	}
 }
 
@@ -1100,15 +1091,9 @@ void CRemoteTreeView::OnMenuDownload(wxCommandEvent& event)
 	CRemoteRecursiveOperation* pRecursiveOperation = m_state.GetRemoteRecursiveOperation();
 	pRecursiveOperation->AddRecursionRoot(std::move(root));
 
-	CServerPath currentPath;
-	const wxTreeItemId selected = GetSelection();
-	if (selected) {
-		currentPath = GetPathFromItem(selected);
-	}
-
 	const bool addOnly = event.GetId() == XRCID("ID_ADDTOQUEUE");
 	CFilterManager filter;
-	pRecursiveOperation->StartRecursiveOperation(CRecursiveOperation::recursive_transfer, filter.GetActiveFilters(), currentPath, !addOnly);
+	pRecursiveOperation->StartRecursiveOperation(recursive_operation::recursive_transfer, filter.GetActiveFilters(), !addOnly);
 }
 
 void CRemoteTreeView::OnMenuDelete(wxCommandEvent&)
@@ -1171,7 +1156,7 @@ void CRemoteTreeView::OnMenuDelete(wxCommandEvent&)
 			m_state.ChangeRemoteDir(startDir);
 		}
 
-		pRecursiveOperation->StartRecursiveOperation(CRecursiveOperation::recursive_delete, filter.GetActiveFilters(), currentPath);
+		pRecursiveOperation->StartRecursiveOperation(recursive_operation::recursive_delete, filter.GetActiveFilters());
 	}
 }
 
@@ -1355,7 +1340,8 @@ CServerPath CRemoteTreeView::MenuMkdir()
 		return CServerPath();
 	}
 
-	m_state.m_pCommandQueue->ProcessCommand(new CMkdirCommand(newPath));
+	transfer_flags const flags = GetMkdirFlags(m_state.GetSite().server, *COptions::Get(), newPath);
+	m_state.m_pCommandQueue->ProcessCommand(new CMkdirCommand(newPath, flags));
 
 	return newPath;
 }
@@ -1369,7 +1355,7 @@ bool CRemoteTreeView::ListExpand(wxTreeItemId item)
 	}
 
 	CDirectoryListing listing;
-	if (m_state.m_pEngine->CacheLookup(path, listing) == FZ_REPLY_OK) {
+	if (m_state.engine_->CacheLookup(path, listing) == FZ_REPLY_OK) {
 		RefreshItem(item, listing, false);
 	}
 	else {
@@ -1431,7 +1417,7 @@ void CRemoteTreeView::ApplyFilters(bool resort)
 		}
 
 		CDirectoryListing listing;
-		if (m_state.m_pEngine->CacheLookup(parent.path, listing) == FZ_REPLY_OK)
+		if (m_state.engine_->CacheLookup(parent.path, listing) == FZ_REPLY_OK)
 			RefreshItem(parent.item, listing, false);
 		else if (filter.HasActiveFilters()) {
 			for (wxTreeItemId child = GetFirstChild(parent.item, cookie); child; child = GetNextSibling(child)) {
@@ -1510,25 +1496,25 @@ void CRemoteTreeView::OnMenuGeturl(wxCommandEvent& event)
 
 void CRemoteTreeView::UpdateSortMode()
 {
-	CFileListCtrlSortBase::NameSortMode sortMode;
-	switch (COptions::Get()->GetOptionVal(OPTION_FILELIST_NAMESORT))
+	NameSortMode sortMode;
+	switch (COptions::Get()->get_int(OPTION_FILELIST_NAMESORT))
 	{
 	case 0:
 	default:
-		sortMode = CFileListCtrlSortBase::namesort_caseinsensitive;
+		sortMode = NameSortMode::case_insensitive;
 		break;
 	case 1:
-		sortMode = CFileListCtrlSortBase::namesort_casesensitive;
+		sortMode = NameSortMode::case_sensitive;
 		break;
 	case 2:
-		sortMode = CFileListCtrlSortBase::namesort_natural;
+		sortMode = NameSortMode::natural;
 		break;
 	}
 	sortFunction_ = CFileListCtrlSortBase::GetCmpFunction(sortMode);
 	Resort();
 }
 
-void CRemoteTreeView::OnOptionsChanged(changed_options_t const& options)
+void CRemoteTreeView::OnOptionsChanged(watched_options const& options)
 {
 	if (options.test(OPTION_FILELIST_NAMESORT)) {
 		UpdateSortMode();
