@@ -1,14 +1,25 @@
-#include <filezilla.h>
+#include "filezilla.h"
 #include "view.h"
 #include "viewheader.h"
 
+#include "state.h"
+#include "listingcomparison.h"
+#include "conditionaldialog.h"
+
+#include <algorithm>
+
 BEGIN_EVENT_TABLE(CView, wxNavigationEnabled<wxWindow>)
 EVT_SIZE(CView::OnSize)
 END_EVENT_TABLE()
 
 CView::CView(wxWindow* pParent)
 {
-	Create(pParent, wxID_ANY, wxDefaultPosition, wxDefaultSize, wxSUNKEN_BORDER);
+#if defined(__WXMAC__) && wxCHECK_VERSION(3, 1, 0)
+	int const border = wxBORDER_NONE;
+#else
+	int const border = wxBORDER_SUNKEN;
+#endif
+	Create(pParent, wxID_ANY, wxDefaultPosition, wxDefaultSize, border);
 }
 
 void CView::SetStatusBar(wxStatusBar* pStatusBar)
@@ -21,6 +32,11 @@ void CView::SetFooter(wxWindow* footer)
 	m_pFooter = footer;
 }
 
+void CView::SetSearchPanel(wxWindow* panel)
+{
+	m_pSearchPanel = panel;
+}
+
 void CView::Arrange(wxWindow* child, wxRect& clientRect, bool top)
 {
 	if (child && child->IsShown()) {
@@ -52,6 +68,7 @@ void CView::OnSize(wxSizeEvent&)
 	Arrange(m_pHeader, rect, true);
 	Arrange(m_pFooter, rect, false);
 	Arrange(m_pStatusBar, rect, false);
+	Arrange(m_pSearchPanel, rect, false);
 
 	if (m_pWnd) {
 		m_pWnd->SetSize(rect);
@@ -85,3 +102,24 @@ CViewHeader* CView::DetachHeader()
 	m_pHeader = 0;
 	return pHeader;
 }
+
+void CView::ShowSearchPanel()
+{
+	if (m_pSearchPanel) {
+		CState* pState = CContextManager::Get()->GetCurrentContext();
+		if (pState) {
+			CComparisonManager* pComparisonManager = pState->GetComparisonManager();
+			if (pComparisonManager && pComparisonManager->IsComparing()) {
+				CConditionalDialog dlg(this, CConditionalDialog::quick_search, CConditionalDialog::yesno);
+				dlg.SetTitle(_("Directory comparison"));
+				dlg.AddText(_("Quick search cannot be opened if comparing directories."));
+				dlg.AddText(_("End comparison and open quick search?"));
+				if (!dlg.Run())
+					return;
+				pComparisonManager->ExitComparisonMode();
+			}		
+		}		
+
+		m_pSearchPanel->Show();
+	}
+}
